def ceil32(n):
    return floor32(n + 31)


def floor32(n):
    return n & ~31


def to_int(stack_item) -> int:
    if isinstance(stack_item, int):
        return stack_item
    return int.from_bytes(stack_item, "big")


def to_bytes(stack_item) -> bytes:
    if isinstance(stack_item, bytes):
        return stack_item.rjust(32, b"\x00")
    return stack_item.to_bytes(32, "big")



import contextlib
import sys

import boa.explorer
from boa.contracts.base_evm_contract import BoaError
from boa.contracts.vyper.vyper_contract import check_boa_error_matches
from boa.dealer import deal
from boa.debugger import BoaDebug
from boa.environment import Env
from boa.explorer import Etherscan, _set_etherscan, get_etherscan
from boa.interpret import (
    from_etherscan,
    load,
    load_abi,
    load_partial,
    load_vyi,
    loads,
    loads_abi,
    loads_partial,
    loads_vyi,
)
from boa.network import NetworkEnv
from boa.precompile import precompile
from boa.test.strategies import fuzz
from boa.util.open_ctx import Open
from boa.verifiers import get_verifier, set_verifier, verify
from boa.vm.py_evm import enable_pyevm_verbose_logging, patch_opcode

# turn off tracebacks if we are in repl
# https://stackoverflow.com/a/64523765
if hasattr(sys, "ps1"):  # pragma: no cover
    pass
    # sys.tracebacklimit = 0

env = Env.get_singleton()


@contextlib.contextmanager
def swap_env(new_env):
    with set_env(new_env):
        yield


def _set_env(new):
    global env
    env = new
    Env._singleton = new


def set_env(new_env):
    global env
    get_env = lambda: env  # noqa: E731
    return Open(get_env, _set_env, new_env)


def fork(
    url: str, block_identifier: int | str = "safe", allow_dirty: bool = False, **kwargs
):
    global env
    if env.evm.is_state_dirty and not allow_dirty:
        raise Exception(
            "Cannot fork with dirty state. Set allow_dirty=True to override."
        )

    new_env = Env()
    new_env.fork(url=url, block_identifier=block_identifier, deprecated=False, **kwargs)
    return set_env(new_env)


def set_browser_env(address=None):
    """Set the environment to use the browser's network in Jupyter/Colab"""
    # import locally because jupyter is generally not installed
    from boa.integrations.jupyter import BrowserEnv

    return set_env(BrowserEnv(address))


def set_network_env(url):
    """Set the environment to use a custom network URL"""
    return set_env(NetworkEnv.from_url(url))


def set_etherscan(*args, **kwargs):
    explorer = Etherscan(*args, **kwargs)
    return Open(get_etherscan, _set_etherscan, explorer)


def reset_env():
    set_env(Env())


def _breakpoint(computation):
    BoaDebug(computation).start()


patch_opcode(0xA6, _breakpoint)


@contextlib.contextmanager
def reverts(*args, **kwargs):
    try:
        yield
        raise ValueError("Did not revert")
    except BoaError as b:
        if args or kwargs:
            check_boa_error_matches(b, *args, **kwargs)


def eval(code):
    return loads("").eval(code)


def _jupyter_server_extension_points() -> list[dict]:
    """
    Returns a list of dictionaries with metadata describing
    where to find the `_load_jupyter_server_extension` function.
    """
    return [{"module": "boa.integrations.jupyter"}]



from dataclasses import dataclass
from typing import Any


@dataclass
class Event:
    log_id: int  # internal py-evm log id, for ordering purposes
    address: str  # checksum address
    event_type: Any  # vyper.semantics.types.user.EventT
    topics: list[Any]  # list of decoded topics
    args: list[Any]  # list of decoded args

    def __repr__(self):
        t_i = 0
        a_i = 0
        b = []
        # align the evm topic + args lists with the way they appear in the source
        # ex. Transfer(indexed address, address, indexed address)
        for is_topic, k in zip(
            self.event_type.indexed, self.event_type.arguments.keys()
        ):
            if is_topic:
                b.append((k, self.topics[t_i]))
                t_i += 1
            else:
                b.append((k, self.args[a_i]))
                a_i += 1

        args = ", ".join(f"{k}={v}" for k, v in b)
        return f"{self.event_type.name}({args})"


@dataclass
class RawEvent:
    event_data: Any



"""
This module implements the BrowserSigner class, which is used to sign transactions
in IPython/JupyterLab/Google Colab.
"""
import json
import logging
import os
from asyncio import get_event_loop, sleep
from itertools import chain
from multiprocessing.shared_memory import SharedMemory
from os import urandom
from os.path import dirname, join, realpath
from typing import Any

import nest_asyncio
from IPython.display import Javascript, display

from boa.integrations.jupyter.constants import (
    ADDRESS_TIMEOUT_MESSAGE,
    CALLBACK_TOKEN_CHARS,
    CALLBACK_TOKEN_TIMEOUT,
    NUL,
    RPC_TIMEOUT_MESSAGE,
    SHARED_MEMORY_LENGTH,
    TRANSACTION_TIMEOUT_MESSAGE,
)
from boa.network import NetworkEnv
from boa.rpc import RPC, RPCError
from boa.util.abi import Address

try:
    from google.colab.output import eval_js as colab_eval_js
except ImportError:
    colab_eval_js = None  # not in Google Colab, use SharedMemory instead


nest_asyncio.apply()


def _install_javascript_triggers():
    """Run the ethers and titanoboa_jupyterlab Javascript snippets in the browser."""
    cur_dir = dirname(realpath(__file__))
    with open(join(cur_dir, "jupyter.js")) as f:
        js = f.read()

    prefix = os.getenv("JUPYTERHUB_SERVICE_PREFIX", "..")
    js = js.replace("$$JUPYTERHUB_SERVICE_PREFIX", prefix)
    js = js.replace("$$BOA_DEBUG_MODE", json.dumps(BrowserRPC._debug_mode))

    display(Javascript(js))


class BrowserRPC(RPC):
    """
    An RPC object that sends requests to the browser via Javascript.
    """

    _debug_mode = False

    def __init__(self):
        if not colab_eval_js:
            # colab creates a new iframe for every call, we need to re-inject it every time
            # for jupyterlab we only need to do it once
            _install_javascript_triggers()

    @property
    def identifier(self) -> str:
        return type(self).__name__  # every instance does the same

    @property
    def name(self):
        return self.identifier

    def fetch(
        self, method: str, params: Any, timeout_message=RPC_TIMEOUT_MESSAGE
    ) -> Any:
        return _javascript_call("rpc", method, params, timeout_message=timeout_message)

    def fetch_multi(
        self, payloads: list[tuple[str, Any]], timeout_message=RPC_TIMEOUT_MESSAGE
    ) -> list[Any]:
        return _javascript_call("multiRpc", payloads, timeout_message=timeout_message)

    def wait_for_tx_receipt(self, tx_hash, timeout: float, poll_latency=1):
        # we do the polling in the browser to avoid too many callbacks
        # each callback generates currently 10px empty space in the frontend
        timeout_ms, pool_latency_ms = timeout * 1000, poll_latency * 1000
        return _javascript_call(
            "waitForTransactionReceipt",
            tx_hash,
            timeout_ms,
            pool_latency_ms,
            timeout_message=RPC_TIMEOUT_MESSAGE,
        )


class BrowserSigner:
    """
    A BrowserSigner is a class that can be used to sign transactions in IPython/JupyterLab.
    """

    def __init__(self, address=None, rpc=None):
        """
        Create a BrowserSigner instance.
        :param address: The account address. If not provided, it will be requested from the browser.
        """
        if rpc is None:
            rpc = BrowserRPC()  # note: the browser window is global anyway
        self._rpc = rpc
        self._given_address = address
        self.address = address

        self.update_address()

    def send_transaction(self, tx_data: dict) -> dict:
        """
        Implements the Account class' send_transaction method.
        It executes a Javascript snippet that requests the user's signature for the transaction.
        Then, it waits for the signature to be received via the API.
        :param tx_data: The transaction data to sign.
        :return: The signed transaction data.
        """
        hash = self._rpc.fetch(
            "eth_sendTransaction", [tx_data], TRANSACTION_TIMEOUT_MESSAGE
        )
        return {"hash": hash}

    def sign_typed_data(self, full_message: dict[str, Any]) -> str:
        """
        Sign typed data value with types data structure for domain using the EIP-712 specification.
        :param full_message: The full message to sign.
        :return: The signature.
        """
        return self._rpc.fetch(
            "eth_signTypedData_v4",
            [self.address, full_message],
            TRANSACTION_TIMEOUT_MESSAGE,
        )

    def update_address(self):
        address = getattr(self._given_address, "address", self._given_address)
        accounts = self._rpc.fetch("eth_requestAccounts", [], ADDRESS_TIMEOUT_MESSAGE)

        if address is None and len(accounts) > 0:
            address = accounts[0]

        if address not in accounts:
            raise ValueError(f"Address {address} is not available in the browser")

        self.address = Address(address)


class BrowserEnv(NetworkEnv):
    """
    A NetworkEnv object that uses the BrowserSigner and BrowserRPC classes.
    """

    _rpc = BrowserRPC()  # Browser is always global anyway, we can make it static

    def __init__(self, address=None, **kwargs):
        super().__init__(self._rpc, **kwargs)
        self.signer = BrowserSigner(address, self._rpc)
        self._update_signer()

    def _update_signer(self):
        self.signer.update_address()
        self.add_account(self.signer, force_eoa=True)

    def execute_code(self, *args, **kwargs):
        self._update_signer()
        return super().execute_code(*args, **kwargs)

    def deploy(self, *args, **kwargs):
        self._update_signer()
        return super().deploy(*args, **kwargs)

    def set_chain_id(self, chain_id: int | str):
        self._rpc.fetch(
            "wallet_switchEthereumChain",
            [{"chainId": chain_id if isinstance(chain_id, str) else hex(chain_id)}],
        )
        self._reset_fork()


def _javascript_call(js_func: str, *args, timeout_message: str) -> Any:
    """
    This function attempts to call a Javascript function in the browser and then
    wait for the result to be sent back to the API.
    - Inside Google Colab, it uses the eval_js function to call the Javascript function.
    - Outside, it uses a SharedMemory object and polls until the frontend called our API.
    A custom timeout message is useful for user feedback.
    :param snippet: A function that given a token and some kwargs, returns a Javascript snippet.
    :param kwargs: The arguments to pass to the Javascript snippet.
    :return: The result of the Javascript snippet sent to the API.
    """
    token = _generate_token()
    args_str = ", ".join(json.dumps(p, cls=_BytesEncoder) for p in chain([token], args))
    js_code = f"window._titanoboa.{js_func}({args_str});"
    if BrowserRPC._debug_mode:
        logging.warning(f"Calling {js_func} with {args_str}")

    if colab_eval_js:
        _install_javascript_triggers()
        result = colab_eval_js(js_code)
        return _parse_js_result(json.loads(result))

    memory = SharedMemory(name=token, create=True, size=SHARED_MEMORY_LENGTH)
    logging.info(f"Waiting for {token}")
    try:
        memory.buf[:1] = NUL
        hide_output_element = "element.style.display = 'none';"
        display(Javascript(js_code + hide_output_element))
        message_bytes = _wait_buffer_set(memory.buf, timeout_message)
        return _parse_js_result(json.loads(message_bytes.decode()))
    finally:
        memory.unlink()  # get rid of the SharedMemory object after it's been used


def _generate_token():
    """Generate a secure unique token to identify the SharedMemory object."""
    return urandom(CALLBACK_TOKEN_CHARS // 2).hex()


def _wait_buffer_set(buffer: memoryview, timeout_message: str) -> bytes:
    """
    Wait for the SharedMemory object to be filled with data.
    :param buffer: The buffer to wait for.
    :param timeout_message: The message to show if the timeout is reached.
    :return: The contents of the buffer.
    """

    async def _async_wait(deadline: float) -> bytes:
        inner_loop = get_event_loop()
        while buffer.tobytes().startswith(NUL):
            if inner_loop.time() > deadline:
                raise TimeoutError(timeout_message)
            await sleep(0.01)

        return buffer.tobytes().split(NUL)[0]

    loop = get_event_loop()
    future = _async_wait(deadline=loop.time() + CALLBACK_TOKEN_TIMEOUT.total_seconds())
    task = loop.create_task(future)
    loop.run_until_complete(task)
    return task.result()


def _parse_js_result(result: dict) -> Any:
    if "data" in result:
        return result["data"]

    def _find_key(input_dict, target_key, typ) -> Any:
        for key, value in input_dict.items():
            if isinstance(value, dict):
                found = _find_key(value, target_key, typ)
                if found is not None:
                    return found
            if key == target_key and isinstance(value, typ) and value != "error":
                return value
        return None

    # raise the error in the Jupyter cell so that the user can see it
    error = result["error"]
    error = error.get("data", error)
    raise RPCError(
        message=_find_key(error, "message", str) or _find_key(error, "error", str),
        code=_find_key(error, "code", int) or -1,
    )


class _BytesEncoder(json.JSONEncoder):
    """
    A JSONEncoder that converts bytes to hex strings to be passed to JavaScript.
    """

    def default(self, o):
        if isinstance(o, bytes):
            return "0x" + o.hex()
        return super().default(o)



import time
from dataclasses import dataclass
from typing import Optional

from boa.rpc import json

try:
    from requests_cache import CachedSession

    SESSION = CachedSession(
        "~/.cache/titanoboa/explorer_cache",
        filter_fn=lambda response: _is_success_response(response.json()),
        allowable_codes=[200],
        cache_control=True,
        expire_after=3600 * 6,
        stale_if_error=True,
        stale_while_revalidate=True,
    )
except ImportError:
    from requests import Session

    SESSION = Session()

DEFAULT_ETHERSCAN_URI = "https://api.etherscan.io/api"


@dataclass
class Etherscan:
    uri: Optional[str] = DEFAULT_ETHERSCAN_URI
    api_key: Optional[str] = None
    num_retries: int = 10
    backoff_ms: int | float = 400.0
    backoff_factor: float = 1.1  # 1.1**10 ~= 2.59

    def __post_init__(self):
        if self.uri is None:
            self.uri = DEFAULT_ETHERSCAN_URI

    def _fetch(self, **params) -> dict:
        """
        Fetch data from Etherscan API.
        Offers a simple caching mechanism to avoid redundant queries.
        Retries if rate limit is reached.
        :param num_retries: Number of retries
        :param backoff_ms: Backoff in milliseconds
        :param params: Additional query parameters
        :return: JSON response
        """
        params = {**params, "apiKey": self.api_key}
        for i in range(self.num_retries):
            res = SESSION.get(self.uri, params=params)
            res.raise_for_status()
            data = res.json()
            if not _is_rate_limited(data):
                break

            f = self.backoff_factor**i
            seconds = self.backoff_ms / 1000
            time.sleep(f * seconds)

        if not _is_success_response(data):
            raise ValueError(f"Failed to retrieve data from API: {data}")

        return data

    def fetch_abi(self, address: str):
        # resolve implementation address if `address` is a proxy contract
        address = self._resolve_implementation_address(address)

        # fetch ABI of `address`
        params = dict(module="contract", action="getabi", address=address)
        data = self._fetch(**params)

        return json.loads(data["result"].strip())

    # fetch the address of a contract; resolves at most one layer of
    # indirection if the address is a proxy contract.
    def _resolve_implementation_address(self, address: str):
        params = dict(module="contract", action="getsourcecode", address=address)
        data = self._fetch(**params)
        source_data = data["result"][0]

        # check if the contract is a proxy
        if int(source_data["Proxy"]) == 1:
            return source_data["Implementation"]
        else:
            return address


_etherscan = Etherscan()


def get_etherscan():
    return _etherscan


def _set_etherscan(etherscan: Etherscan):
    global _etherscan
    _etherscan = etherscan


def _is_success_response(data: dict) -> bool:
    return data.get("status") == "1"


def _is_rate_limited(data: dict) -> bool:
    """
    Check if the response is rate limited. Possible error messages:
    - Max calls per sec rate limit reached (X/sec)
    - Max rate limit reached, please use API Key for higher rate limit
    - Max rate limit reached
    :param data: Etherscan API response
    :return: True if rate limited, False otherwise
    """
    return "rate limit" in data.get("result", "") and data.get("status") == "0"



import json
import sqlite3
import uuid
from dataclasses import asdict, dataclass, field, fields
from pathlib import Path
from typing import Any, Generator, Optional

from boa.util.abi import Address
from boa.util.open_ctx import Open

"""
Module to handle deployment objects. When a contract is deployed in network
mode, we enter it into the deployments database so that it can be
queried/verified later.

This module could potentially be handled as plugin functionality / or left
as functionality for higher-level frameworks.
"""

_session_id: str = None  # type: ignore


# generate a unique session id, so that deployments can be queried by session
def get_session_id():
    global _session_id
    if _session_id is None:
        _session_id = str(uuid.uuid4())
    return _session_id


@dataclass(frozen=True)
class Deployment:
    contract_address: Address  # receipt_dict["createAddress"]
    contract_name: str
    filename: str
    rpc: str
    deployer: Address  # ostensibly equal to tx_dict["from"]
    tx_hash: str
    broadcast_ts: float  # time the tx was broadcast
    tx_dict: dict  # raw tx fields
    receipt_dict: dict  # raw receipt fields
    source_code: Optional[Any]  # optional source code or bundle
    abi: Optional[Any]
    session_id: str = field(default_factory=get_session_id)
    deployment_id: Optional[int] = None  # the db-assigned id - primary key

    def sql_values(self):
        ret = asdict(self)
        # sqlite doesn't have json, just dump to string
        ret["tx_dict"] = json.dumps(ret["tx_dict"])
        ret["receipt_dict"] = json.dumps(ret["receipt_dict"])
        if ret["source_code"] is not None:
            ret["source_code"] = json.dumps(ret["source_code"])
        if ret["abi"] is not None:
            ret["abi"] = json.dumps(ret["abi"])
        return ret

    def to_dict(self):
        """
        Convert Deployment object to a dict, which is prepared to be
        dumped to json.
        """
        return asdict(self)

    def to_json(self, *args, **kwargs):
        """
        Convert a Deployment object to a json object. *args and **kwargs
        are forwarded to the `json.dumps()` call.
        """
        return json.dumps(self.to_dict(), *args, **kwargs)

    @classmethod
    def from_sql_tuple(cls, values):
        assert len(values) == len(fields(cls))
        ret = dict(zip([field.name for field in fields(cls)], values))
        ret["contract_address"] = Address(ret["contract_address"])
        ret["deployer"] = Address(ret["deployer"])
        ret["tx_dict"] = json.loads(ret["tx_dict"])
        ret["receipt_dict"] = json.loads(ret["receipt_dict"])
        if ret["source_code"] is not None:
            ret["source_code"] = json.loads(ret["source_code"])
        if ret["abi"] is not None:
            ret["abi"] = json.loads(ret["abi"])
        return cls(**ret)


_CREATE_CMD = """
CREATE TABLE IF NOT EXISTS
    deployments(
        deployment_id integer primary key autoincrement,
        session_id text,
        contract_address text,
        contract_name text,
        filename text,
        rpc text,
        deployer text,
        tx_hash text,
        broadcast_ts real,
        tx_dict text,
        receipt_dict text,
        source_code text,
        abi text
    );
"""


class DeploymentsDB:
    def __init__(self, path=":memory:"):
        if path != ":memory:":  # sqlite magic path
            path = Path(path)
            path.parent.mkdir(parents=True, exist_ok=True)

        # once 3.12 is min version, use autocommit=True
        self.db = sqlite3.connect(path)
        self.db.execute(_CREATE_CMD)

        # Migration for legacy DB without filename column
        # We can/should remove this after some time (all 4 users migrate)
        self._apply_filename_migration()

    def __del__(self):
        self.db.close()

    def insert_deployment(self, deployment: Deployment):
        values = deployment.sql_values()

        values_placeholder = ",".join(["?"] * len(values))
        colnames = ",".join(values.keys())

        insert_cmd = f"INSERT INTO deployments({colnames}) VALUES({values_placeholder})"

        self.db.execute(insert_cmd, tuple(values.values()))
        self.db.commit()

    def _apply_filename_migration(self) -> None:
        cursor = self.db.execute("PRAGMA table_info(deployments);")
        columns = [col[1] for col in cursor.fetchall()]
        is_in_db = "filename" in columns
        if is_in_db:
            return
        self.db.execute("ALTER TABLE deployments ADD COLUMN filename text;")

    def _get_deployments_from_sql(self, sql_query: str, parameters=(), /):
        cur = self.db.execute(sql_query, parameters)
        return (Deployment.from_sql_tuple(item) for item in cur)

    def _get_fieldnames_str(self) -> str:
        return ",".join(field.name for field in fields(Deployment))

    def get_deployments(self) -> Generator[Deployment, None, None]:
        """
        Return all the deployments from the database. Returns an iterator
        which can be converted to a list with `list(db.get_deployments())`.

        Returns the deployments ordered by most recent first, i.e. using
        an `ORDER BY deployment_id DESC` clause.
        """
        fieldnames = self._get_fieldnames_str()
        return self._get_deployments_from_sql(
            f"SELECT {fieldnames} FROM deployments order by deployment_id desc"
        )


_db: Optional[DeploymentsDB] = None


def set_deployments_db(db: Optional[DeploymentsDB]):
    def set_(db):
        global _db
        _db = db

    return Open(get_deployments_db, set_, db)


def get_deployments_db():
    global _db
    return _db



class lrudict(dict):
    def __init__(self, n, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n = n

    def __getitem__(self, k):
        val = super().__getitem__(k)
        del self[k]  # move to the front of the queue
        super().__setitem__(k, val)
        return val

    def __setitem__(self, k, val):
        if len(self) == self.n and k not in self:
            del self[next(iter(self))]
        super().__setitem__(k, val)

    # set based on a lambda
    def setdefault_lambda(self, k, fn):
        try:
            return self[k]
        except KeyError:
            self[k] = (ret := fn(k))
            return ret



from functools import cached_property

import coverage.plugin
import vyper.ast as vy_ast
from vyper.ast.parse import parse_to_ast

from boa.contracts.vyper.ast_utils import get_fn_ancestor_from_node
from boa.environment import Env


def coverage_init(registry, options):
    plugin = TitanoboaPlugin(options)
    registry.add_file_tracer(plugin)
    registry.add_configurer(plugin)

    # set on the class so that reset_env() doesn't disable tracing
    Env._coverage_enabled = True


class TitanoboaPlugin(coverage.plugin.CoveragePlugin):
    def __init__(self, options):
        pass

    def configure(self, config):
        config.get_option("run:source_pkgs").append("boa.environment")

    def file_tracer(self, filename):
        if filename.endswith("boa/environment.py"):
            return TitanoboaTracer()

    def file_reporter(self, filename):
        if filename.endswith(".vy"):
            return TitanoboaReporter(filename)


class TitanoboaTracer(coverage.plugin.FileTracer):
    def __init__(self):
        pass

    # coverage.py requires us to inspect the python call frame to
    # see what line number to produce. we hook into specially crafted
    # Env._trace_cov which is called for every unique pc if coverage is
    # enabled, and then back out the contract and lineno information
    # from there.

    def _valid_frame(self, frame):
        if hasattr(frame.f_code, "co_qualname"):
            # Python>=3.11
            code_qualname = frame.f_code.co_qualname
            return code_qualname == Env._trace_cov.__qualname__

        else:
            # in Python<3.11 we don't have co_qualname, so try hard to
            # find a match anyways. (this might fail if for some reason
            # the executing env has a monkey-patched _trace_cov
            # or something)
            env = Env.get_singleton()
            return frame.f_code == env._trace_cov.__code__

    def dynamic_source_filename(self, filename, frame):
        if not self._valid_frame(frame):
            return None
        ret = frame.f_locals["filename"]
        if ret is not None and not ret.endswith(".vy"):
            return None
        return ret

    def has_dynamic_source_filename(self):
        return True

    # https://coverage.rtfd.io/en/stable/api_plugin.html#coverage.FileTracer.line_number_range
    def line_number_range(self, frame):
        if not self._valid_frame(frame):
            return (-1, -1)

        ast_node = frame.f_locals["node"]

        start_lineno = ast_node.lineno
        end_lineno = ast_node.end_lineno
        if end_lineno is None:
            end_lineno = start_lineno
        return start_lineno, end_lineno

    # XXX: dynamic context. return function name or something
    def dynamic_context(self, frame):
        pass


# helper function. null returns get optimized directly into a jump
# to function cleanup which maps to the parent FunctionDef ast.
def _is_null_return(ast_node):
    match ast_node:
        case vy_ast.Return(value=None):
            return True
    return False


class TitanoboaReporter(coverage.plugin.FileReporter):
    def __init__(self, filename, env=None):
        super().__init__(filename)

    @cached_property
    def _ast(self):
        return parse_to_ast(self.source())

    def arcs(self):
        ret = set()

        for ast_node in self._ast.get_descendants(vy_ast.If):
            fn_node = get_fn_ancestor_from_node(ast_node)

            # one arc is directly into the body
            arc_true = ast_node.body[0].lineno
            if _is_null_return(ast_node.body[0]):
                arc_true = fn_node.lineno
            ret.add((ast_node.lineno, arc_true))

            # the other arc is to the end of the if statement
            # try hard to find the next executable line.
            children = ast_node._parent.get_children()
            for node, next_ in zip(children, children[1:]):
                if id(node) == id(ast_node):
                    arc_false = next_.lineno
                    break
            else:
                # the if stmt was the last stmt in the enclosing scope.
                arc_false = ast_node._parent.end_lineno + 1

            # unless there is an else or elif. then the other
            # arc is to the else/elif statement.
            if ast_node.orelse:
                arc_false = ast_node.orelse[0].lineno

            # return cases:
            # if it's past the end of the fn it's an implicit return
            if arc_false > fn_node.end_lineno:
                arc_false = fn_node.lineno
            # or it's an explicit return
            if ast_node.orelse and _is_null_return(ast_node.orelse[0]):
                arc_false = fn_node.lineno

            ret.add((ast_node.lineno, arc_false))

        return ret

    def exit_counts(self):
        ret = {}
        for ast_node in self._ast.get_descendants(vy_ast.If):
            ret[ast_node.lineno] = 2
        return ret

    @cached_property
    def _lines(self):
        ret = set()

        functions = self._ast.get_children(vy_ast.FunctionDef)

        for f in functions:
            for stmt in f.body:
                ret.add(stmt.lineno)
                for node in stmt.get_descendants():
                    if isinstance(node, vy_ast.AnnAssign) and isinstance(
                        node.parent, vy_ast.For
                    ):
                        # tokenizer bug with vyper parser, just ignore it
                        continue
                    ret.add(node.lineno)

        return ret

    # OVERRIDES
    def lines(self):
        return self._lines



import textwrap

import vyper.ast as vy_ast
import vyper.semantics.analysis as analysis
from vyper.ast.parse import parse_to_ast
from vyper.codegen.function_definitions import (
    generate_ir_for_external_function,
    generate_ir_for_internal_function,
)
from vyper.codegen.ir_node import IRnode
from vyper.codegen.module import _runtime_reachable_functions
from vyper.compiler.settings import anchor_settings
from vyper.exceptions import InvalidType
from vyper.ir import compile_ir, optimizer
from vyper.semantics.analysis.constant_folding import ConstantFolder
from vyper.semantics.analysis.utils import get_exact_type_from_node

from boa.contracts.vyper.ir_executor import executor_from_ir

# id used internally for method id name
_METHOD_ID_VAR = "_calldata_method_id"


# visit dst_ast with the constants of src_ast. because of the way we
# construct the analysis, we don't insert most of the original contract
# into the vyper_function, relying on semantic data already being in the
# namespace. however, as of 0.4.0, constant folding doesn't rely on namespace,
# all the constants are handled by data structures internal to ConstantFolder.
# here we visit dst_ast with the constants of src_ast.
def _swipe_constants(src_ast, dst_ast):
    s = ConstantFolder(src_ast)
    s._get_constants()
    s.visit(dst_ast)


def compile_vyper_function(vyper_function, contract):
    """Compiles a vyper function and appends it to the top of the IR of a
    contract. This is useful for vyper `eval` and internal functions, where
    the runtime bytecode must be changed to add more runtime functionality
    (such as eval, and calling internal functions)
    (performance note: this function is very very slow!)
    """

    compiler_data = contract.compiler_data

    with anchor_settings(compiler_data.settings):
        module_t = contract.module_t
        ast = parse_to_ast(vyper_function)

        # override namespace and add wrapper code at the top
        with contract.override_vyper_namespace():
            _swipe_constants(compiler_data.annotated_vyper_module, ast)
            analysis.analyze_module(ast, compiler_data.input_bundle)

        ast = ast.body[0]
        func_t = ast._metadata["func_type"]

        contract.ensure_id(func_t)
        funcinfo = generate_ir_for_external_function(ast, module_t)
        ir = funcinfo.common_ir

        entry_label = func_t._ir_info.external_function_base_entry_label

        ir = ["seq", ["goto", entry_label], ir]

        # use a dummy method id
        ir = ["with", _METHOD_ID_VAR, 0, ir]

        # first mush it with the rest of the IR in the contract to ensure
        # all labels are present, and then optimize all together
        # (use unoptimized IR, ir_executor can't handle optimized selector tables)
        _, contract_runtime = contract.unoptimized_ir
        ir_list = ["seq", ir, contract_runtime]
        reachable = func_t.reachable_internal_functions

        already_compiled = _runtime_reachable_functions(module_t, contract)
        missing_functions = reachable.difference(already_compiled)
        # TODO: cache function compilations or something
        for f in missing_functions:
            assert f.ast_def is not None
            contract.ensure_id(f)
            ir_list.append(
                generate_ir_for_internal_function(f.ast_def, module_t, False).func_ir
            )

        ir = IRnode.from_list(ir_list)
        ir = optimizer.optimize(ir)

        assembly = compile_ir.compile_to_assembly(ir)
        bytecode, source_map = compile_ir.assembly_to_evm(assembly)
        bytecode += contract.data_section
        typ = func_t.return_type

        # generate the IR executor
        ir_executor = executor_from_ir(ir, compiler_data)

        return ast, ir_executor, bytecode, source_map, typ


def generate_bytecode_for_internal_fn(fn):
    """Wraps internal fns with an external fn and generated bytecode"""
    contract = fn.contract
    fn_name = fn.func_t.name
    fn_ast = fn.func_t.ast_def

    fn_args = ", ".join([arg.name for arg in fn.func_t.arguments])

    return_sig = ""
    fn_call = ""
    if fn.func_t.return_type:
        return_sig = f" -> {fn_ast.returns.node_source_code}"
        fn_call = "return "
    fn_call += f"self.{fn_name}({fn_args})"

    if len(fn.func_t.arguments) == 0:
        # special case - when there are no arguments, node_source_code is
        # garbage
        fn_sig = ""
    else:
        fn_sig = fn_ast.args.node_source_code.replace("\n", " ")

    wrapper_code = f"""
@external
@payable
def __boa_private_{fn_name}__({fn_sig}){return_sig}:
    {fn_call}
    """
    return compile_vyper_function(wrapper_code, contract)


def generate_bytecode_for_arbitrary_stmt(source_code, contract):
    """Wraps arbitrary stmts with external fn and generates bytecode"""

    ast = parse_to_ast(source_code)

    ast = ast.body[0]

    return_sig = ""
    debug_body = source_code

    if isinstance(ast, vy_ast.Expr):
        with contract.override_vyper_namespace():
            try:
                ast_typ = get_exact_type_from_node(ast.value)
                return_sig = f"-> {ast_typ}"
                debug_body = f"return {source_code}"
            except InvalidType:
                pass

    # wrap code in function so that we can easily generate code for it
    wrapper_code = textwrap.dedent(
        f"""
        @external
        @payable
        def __boa_debug__() {return_sig}:
            {debug_body}
    """
    )
    return compile_vyper_function(wrapper_code, contract)



# Simple context manager which functions like the `open()` builtin -
# if simply called, it never calls __exit__, but if used as a context manager,
# it calls __exit__ at scope exit
class Open:
    def __init__(self, get, set_, item):
        self.anchor = get()
        self._set = set_
        self._set(item)
        self._item = item

    def __enter__(self):
        return self._item

    def __exit__(self, *args):
        self._set(self.anchor)



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from eth.abc import ComputationAPI

from boa.contracts.call_trace import TraceFrame
from boa.environment import Env
from boa.util.abi import Address
from boa.util.exceptions import strip_internal_frames

if TYPE_CHECKING:
    from boa.contracts.vyper.vyper_contract import DevReason
    from boa.vm.py_evm import titanoboa_computation


class _BaseEVMContract:
    """
    Base class for EVM (Ethereum Virtual Machine) contract:
    This includes ABI and Vyper contract.
    """

    # flag to signal whether this contract can be line profiled
    _can_line_profile = False

    def __init__(
        self,
        name: str,
        env: Optional[Env] = None,
        filename: Optional[str] = None,
        address: Optional[Address] = None,
    ):
        self.env = env or Env.get_singleton()
        self.contract_name = name
        self._address = address  # this can be overridden by subclasses
        self.filename = filename
        self._computation: Optional[ComputationAPI] = None

    def stack_trace(self, computation: ComputationAPI):  # pragma: no cover
        raise NotImplementedError

    def call_trace(self) -> TraceFrame:
        assert self._computation is not None, "No computation to trace"
        return self._computation.call_trace

    def handle_error(self, computation):
        try:
            raise BoaError.create(computation, self)
        except BoaError as b:
            # modify the error so the traceback starts in userland.
            # inspired by answers in https://stackoverflow.com/q/1603940/
            raise strip_internal_frames(b) from b

    @property
    def address(self) -> Address:
        if self._address is None:
            # avoid assert, in pytest it would call repr(self) which segfaults
            raise RuntimeError("Contract address is not set")
        return self._address


class StackTrace(list):  # list[str|ErrorDetail]
    def __str__(self):
        return "\n\n".join(str(x) for x in self)

    @property
    def dev_reason(self) -> Optional["DevReason"]:
        if self.last_frame is None or isinstance(self.last_frame, str):
            return None
        return self.last_frame.dev_reason

    @property
    def last_frame(self):
        return self[-1]


def _trace_for_unknown_contract(computation, env):
    err = f"   <Unknown contract 0x{computation.msg.code_address.hex()}>"
    trace = StackTrace([err])
    return _handle_child_trace(computation, env, trace)


def _handle_child_trace(computation, env, return_trace):
    if len(computation.children) == 0:
        return return_trace
    if not computation.children[-1].is_error:
        return return_trace
    child = computation.children[-1]

    # TODO: maybe should be:
    # child_obj = (
    #   env.lookup_contract(child.msg.code_address)
    #   or env._code_registry.get(child.msg.code)
    # )
    child_obj = env.lookup_contract(child.msg.code_address)

    if child_obj is None:
        child_trace = _trace_for_unknown_contract(child, env)
    else:
        child_trace = child_obj.stack_trace(child)

    if child_trace.dev_reason is not None and return_trace.dev_reason is None:
        # Propagate the dev reason from the child frame to the parent
        return_trace.last_frame.dev_reason = child_trace.dev_reason

    return StackTrace(child_trace + return_trace)


@dataclass
class BoaError(Exception):
    call_trace: TraceFrame
    stack_trace: StackTrace

    @classmethod
    def create(cls, computation: "titanoboa_computation", contract: _BaseEVMContract):
        return cls(computation.call_trace, contract.stack_trace(computation))

    # perf TODO: don't materialize the stack trace until we need it,
    # i.e. BoaError ctor only takes what is necessary to construct the
    # stack trace but does not require the actual stack trace itself.
    def __str__(self):
        frame = self.stack_trace.last_frame
        if hasattr(frame, "vm_error"):
            err = frame.vm_error
            if not getattr(err, "_already_pretty", False):
                # avoid double patching when str() is called more than once
                setattr(err, "_already_pretty", True)
                err.args = (frame.pretty_vm_reason, *err.args[1:])
        else:
            err = frame

        ret = f"{err}\n\n{self.stack_trace}"
        call_tree = str(self.call_trace)
        ledge = "=" * 72
        return f"\n{ledge}\n{call_tree}\n{ledge}\n\n{ret}"



import contextlib
from typing import Any

import boa
from boa.util.abi import Address
from boa.vm.utils import to_int

SLOAD_OPCODE = 0x54


class SloadTracer:
    def __init__(self, super_fn):
        self.super_fn = super_fn
        self.trace = []

    def __call__(self, computation):
        slot = to_int(computation._stack.values[-1])
        self.super_fn(computation)
        self.trace.append(slot)


@contextlib.contextmanager
def patch_sload():
    tmp = boa.vm.py_evm._opcode_overrides

    try:
        # clean opcode overrides in case there is user-stuff in there - we
        # don't want to trip user-overridden opcodes, since this is a
        # "system" operation.
        boa.vm.py_evm._opcode_overrides = {}
        opcodes = boa.env.evm.vm.state.computation_class.opcodes
        sload_tracer = SloadTracer(opcodes[SLOAD_OPCODE])
        boa.patch_opcode(SLOAD_OPCODE, sload_tracer)

        yield sload_tracer

    finally:
        # restore
        boa.vm.py_evm._opcode_overrides = tmp


def _get_func(contract, fn_name):
    try:
        return getattr(contract, fn_name)
    except AttributeError:
        raise ValueError(f"Function {fn_name} not found in {contract}")


def update_storage_slot(contract, fn_name, mk_new_value: Any, args):
    # TODO: we should really protect all user-facing trace data-structures in
    # this function, including profiling/gas tracing, sstore+sha3tracer, etc.
    fn = _get_func(contract, fn_name)
    with patch_sload() as sload_tracer:
        result = fn(*args)
        trace = sload_tracer.trace

    # we iterate over all the storage slots accessed by the SLOAD opcode
    # until we find the one that contains the result of the function call
    for slot in trace:
        slot_value = boa.env.get_storage(contract.address, slot)

        # perf: don't bother checking the slot if it doesn't match the expected result
        if slot_value != result:
            continue

        # double check we have the right slot -- modifying the value should
        # yield the expected value on calling the view function again
        with boa.env.anchor():
            poison = (result + 1) % 2**256
            boa.env.set_storage(contract.address, slot, poison)
            if fn(*args) != poison:
                continue

        # we found the slot. update it
        new_value = mk_new_value(slot_value)
        boa.env.set_storage(contract.address, slot, new_value)

        # sanity check
        if fn(*args) != new_value:
            # very unlikely, but this means the ERC20 implementation is totally weird
            raise RuntimeError("new value does not match, unknown ERC20 implementation")

        return result

    msg = f"Could not find the target slot for {fn_name}"
    msg += ", this is expected if the token packs storage slots or"
    msg += " computes the value on the fly"
    raise ValueError(msg)


def deal(token, receiver: Address, amount: int, adjust_supply: bool = True):
    """
    Modifies `receiver` balance of `token` to `amount`, and adjusts the total supply if
    `adjust_supply` is True.
    Inspired by https://github.com/foundry-rs/forge-std/blob/07263d193d/src/StdCheats.sol#L728
    """
    update_balance = lambda _: amount  # noqa: E731
    # we could handle "exotic" tokens if we accepted an additional argument, "expect_amount"
    # so that user with low-level knowledge of how the token works can call deal() with the
    # exact value of the storage slot and instead of checking `balanceOf` returns `amount`,
    # we check that balanceOf returns `expect_amount`.
    old_balance = update_storage_slot(token, "balanceOf", update_balance, (receiver,))

    if adjust_supply:
        delta = amount - old_balance
        update_supply = lambda supply: supply + delta  # noqa: E731
        update_storage_slot(token, "totalSupply", update_supply, ())



from functools import cached_property

from boa.contracts.abi.abi_contract import ABIContractFactory, ABIFunction
from boa.environment import Env
from boa.util.eip5202 import generate_blueprint_bytecode


class VVMDeployer:
    """
    A deployer that uses the Vyper Version Manager (VVM).
    This allows deployment of contracts written in older versions of Vyper that
    can interact with new versions using the ABI definition.
    """

    def __init__(self, abi, bytecode, filename):
        """
        Initialize a VVMDeployer instance.
        :param abi: The contract's ABI.
        :param bytecode: The contract's bytecode.
        :param filename: The filename of the contract.
        """
        self.abi = abi
        self.bytecode = bytecode
        self.filename = filename

    @classmethod
    def from_compiler_output(cls, compiler_output, filename):
        abi = compiler_output["abi"]
        bytecode_nibbles = compiler_output["bytecode"]
        bytecode = bytes.fromhex(bytecode_nibbles.removeprefix("0x"))
        return cls(abi, bytecode, filename)

    @cached_property
    def factory(self):
        return ABIContractFactory.from_abi_dict(self.abi)

    @cached_property
    def constructor(self):
        for t in self.abi:
            if t["type"] == "constructor":
                return ABIFunction(t, contract_name=self.filename)
        return None

    def deploy(self, *args, contract_name=None, env=None, **kwargs):
        encoded_args = b""
        if self.constructor is not None:
            encoded_args = self.constructor.prepare_calldata(*args)
        elif len(args) > 0:
            raise ValueError(f"No constructor, but args were provided: {args}")

        if env is None:
            env = Env.get_singleton()

        address, _ = env.deploy_code(bytecode=self.bytecode + encoded_args, **kwargs)

        # TODO: pass thru contract_name
        return self.at(address)

    @cached_property
    def _blueprint_deployer(self):
        # TODO: add filename
        return ABIContractFactory.from_abi_dict([])

    def deploy_as_blueprint(self, env=None, blueprint_preamble=None, **kwargs):
        """
        Deploy a new blueprint from this contract.
        :param blueprint_preamble: The preamble to use for the blueprint.
        :param env: The environment to deploy the blueprint in.
        :param kwargs: Keyword arguments to pass to the environment `deploy_code` method.
        :returns: A contract instance.
        """
        if env is None:
            env = Env.get_singleton()

        blueprint_bytecode = generate_blueprint_bytecode(
            self.bytecode, blueprint_preamble
        )
        address, _ = env.deploy_code(bytecode=blueprint_bytecode, **kwargs)

        ret = self._blueprint_deployer.at(address)

        env.register_blueprint(self.bytecode, ret)
        return ret

    def __call__(self, *args, **kwargs):
        return self.deploy(*args, **kwargs)

    def at(self, address):
        return self.factory.at(address)



from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from pathlib import Path
from typing import Optional

from eth.abc import ComputationAPI

from boa.rpc import json
from boa.util.abi import Address, abi_decode


class TraceSource:
    def format(self, input_: bytes, output: bytes, is_error: bool):
        in_ = self._format_input(input_)

        if is_error:
            out = self._format_error(output)
            return f"{self}{in_} <{out}>"

        out = self._format_output(output)
        return f"{self}{in_} => {out}"

    def _format_input(self, input_: bytes):
        decoded = abi_decode(self.args_abi_type, input_)
        args = [
            f"{name} = {_to_str(d)}" for d, name in zip(decoded, self._argument_names)
        ]
        return f"({', '.join(args)})"

    def _format_error(self, error_bytes: bytes):
        # b"\x08\xc3y\xa0" == method_id("Error(string)")
        if error_bytes.startswith(b"\x08\xc3y\xa0"):
            (ret,) = abi_decode("(string)", error_bytes[4:])
            return ret

        # TODO: handle other error types
        return "0x" + error_bytes.hex()

    def _format_output(self, output: bytes):
        if output == b"":
            return "0x"

        decoded = abi_decode(self.return_abi_type, output)
        return _to_str(decoded)

    @property
    def args_abi_type(self) -> str:  # must be implemented by subclasses
        raise NotImplementedError  # pragma: no cover

    @property
    def _argument_names(self) -> list[str]:  # must be implemented by subclasses
        raise NotImplementedError  # pragma: no cover

    @property
    def return_abi_type(self) -> str:  # must be implemented by subclasses
        raise NotImplementedError  # pragma: no cover

    def __str__(self):  # must be implemented by subclasses
        raise NotImplementedError  # pragma: no cover


@dataclass
class TraceFrame:
    computation: "ComputationAPI"
    source: Optional[TraceSource]
    depth: int
    children: list["TraceFrame"]

    @cached_property
    def address(self) -> Address:
        if self.computation.msg.is_create:
            return Address(self.computation.msg.storage_address)
        return Address(self.computation.msg.code_address)

    @cached_property
    def gas_used(self) -> int:
        return self.computation.net_gas_used

    @cached_property
    def input_data(self) -> bytes:
        return self.computation.msg.data[4:]

    @cached_property
    def selector(self) -> bytes:
        return self.computation.msg.data[:4]

    @cached_property
    def output(self) -> bytes:
        return self.computation.output

    @cached_property
    def is_error(self) -> bool:
        return self.computation.is_error

    def __str__(self):
        text = f"{' ' * self.depth * 4}{self.text}"
        return "\n".join(chain((text,), (str(child) for child in self.children)))

    @property
    def text(self):
        if self.source:
            text = self.source.format(self.input_data, self.output, self.is_error)
        else:
            text = f"Unknown contract {self.address}"
            if self.computation.msg.data != b"":
                text += ".0x" + self.selector.hex()

        ret = f"[{self.gas_used}] {text}"
        if self.is_error:
            ret = f"[E] {ret}"
        return ret

    def to_dict(self) -> dict:
        return {
            "address": str(self.address),
            "depth": self.depth,
            "gas_used": self.gas_used,
            "source": str(self.source),
            "input": "0x" + self.input_data.hex(),
            "output": "0x" + self.output.hex(),
            "children": [child.to_dict() for child in self.children],
            "is_error": self.is_error,
            "text": self.text,
        }

    def export_html(self, destination: str | Path):
        html = self.to_html()
        with open(destination, "w") as f:
            f.write(html)
        print(f"Trace written to file://{Path(destination).absolute()}")

    def to_html(self):
        with open(Path(__file__).parent / "trace-template.html") as f:
            template = f.read()
        trace_json = json.dumps(self.to_dict()).replace("\\", "\\\\")
        return template.replace("$$TRACE", trace_json)


def _to_str(d):
    if isinstance(d, bytes):
        return "0x" + d.hex()
    if isinstance(d, list):
        return f"[{', '.join(_to_str(x) for x in d)}]"
    if isinstance(d, tuple):
        return f"({', '.join(_to_str(x) for x in d)})"
    if isinstance(d, str):
        return f'"{d}"'
    return str(d)



# the main "entry point" of vyper-related functionality like
# AST handling, traceback construction and ABI (marshaling
# and unmarshaling vyper objects)

import contextlib
import copy
import warnings
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Optional

import vyper
import vyper.ast as vy_ast
import vyper.ir.compile_ir as compile_ir
import vyper.semantics.namespace as vy_ns
from eth.exceptions import VMError
from vyper.ast.nodes import VariableDecl
from vyper.ast.parse import parse_to_ast
from vyper.codegen.core import calculate_type_for_external_return
from vyper.codegen.function_definitions import (
    generate_ir_for_external_function,
    generate_ir_for_internal_function,
)
from vyper.codegen.ir_node import IRnode
from vyper.codegen.module import generate_ir_for_module
from vyper.compiler import CompilerData
from vyper.compiler import output as compiler_output
from vyper.compiler.output import build_abi_output, build_solc_json
from vyper.compiler.settings import OptimizationLevel, anchor_settings
from vyper.exceptions import VyperException
from vyper.ir.optimizer import optimize
from vyper.semantics.types import AddressT, HashMapT, TupleT
from vyper.utils import method_id

from boa import BoaError
from boa.contracts.base_evm_contract import (
    StackTrace,
    _BaseEVMContract,
    _handle_child_trace,
)
from boa.contracts.call_trace import TraceSource
from boa.contracts.vyper.ast_utils import get_fn_ancestor_from_node, reason_at
from boa.contracts.vyper.compiler_utils import (
    _METHOD_ID_VAR,
    compile_vyper_function,
    generate_bytecode_for_arbitrary_stmt,
    generate_bytecode_for_internal_fn,
)
from boa.contracts.vyper.decoder_utils import (
    ByteAddressableStorage,
    decode_vyper_object,
)
from boa.contracts.vyper.event import Event, RawEvent
from boa.contracts.vyper.ir_executor import executor_from_ir
from boa.environment import Env
from boa.profiling import cache_gas_used_for_computation
from boa.util.abi import Address, abi_decode, abi_encode
from boa.util.eip5202 import generate_blueprint_bytecode
from boa.util.lrudict import lrudict
from boa.vm.gas_meters import ProfilingGasMeter
from boa.vm.utils import to_bytes, to_int

# error messages for external calls
EXTERNAL_CALL_ERRORS = ("external call failed", "returndatasize too small")

CREATE_ERRORS = ("create failed", "create2 failed")

# error detail where user possibly provided dev revert reason
DEV_REASON_ALLOWED = ("user raise", "user assert")


class VyperDeployer:
    create_compiler_data = CompilerData  # this may be a different class in plugins

    def __init__(self, compiler_data, filename=None):
        self.compiler_data = compiler_data

        # force compilation so that if there are any errors in the contract,
        # we fail at load rather than at deploy time.
        with anchor_settings(self.compiler_data.settings):
            _ = compiler_data.bytecode, compiler_data.bytecode_runtime

        self.filename = filename

    def __call__(self, *args, **kwargs):
        return self.deploy(*args, **kwargs)

    def deploy(self, *args, **kwargs):
        return VyperContract(
            self.compiler_data, *args, filename=self.filename, **kwargs
        )

    def deploy_as_blueprint(self, *args, **kwargs):
        return VyperBlueprint(
            self.compiler_data, *args, filename=self.filename, **kwargs
        )

    def stomp(self, address: Any, data_section=None) -> "VyperContract":
        address = Address(address)

        ret = self.deploy(override_address=address, skip_initcode=True)
        vm = ret.env.evm
        old_bytecode = vm.get_code(address)
        new_bytecode = self.compiler_data.bytecode_runtime

        immutables_size = self.compiler_data.global_ctx.immutable_section_bytes
        if immutables_size > 0:
            data_section = old_bytecode[-immutables_size:]
            new_bytecode += data_section

        vm.set_code(address, new_bytecode)
        ret.env.register_contract(address, ret)
        ret._set_bytecode(new_bytecode)
        return ret

    # TODO: allow `env=` kwargs and so on
    def at(self, address: Any) -> "VyperContract":
        address = Address(address)

        ret = self.deploy(override_address=address, skip_initcode=True)
        bytecode = ret.env.get_code(address)

        ret._set_bytecode(bytecode)

        ret.env.register_contract(address, ret)
        return ret

    @cached_property
    def solc_json(self):
        """
        Generates a solc "standard json" representation of the Vyper contract.
        """
        return build_solc_json(self.compiler_data)

    @cached_property
    def _constants(self):
        # Make constants available at compile time. Useful for testing. See #196
        return ConstantsModel(self.compiler_data)


# a few lines of shared code between VyperBlueprint and VyperContract
class _BaseVyperContract(_BaseEVMContract):
    def __init__(
        self,
        compiler_data: CompilerData,
        contract_name: Optional[str] = None,
        env: Optional[Env] = None,
        filename: Optional[str] = None,
    ):
        if contract_name is None:
            contract_name = Path(compiler_data.contract_path).stem

        super().__init__(contract_name, env, filename)
        self.compiler_data = compiler_data

        with anchor_settings(self.compiler_data.settings):
            _ = compiler_data.bytecode, compiler_data.bytecode_runtime

        if (capabilities := getattr(env, "capabilities", None)) is not None:
            compiler_evm_version = self.compiler_data.settings.evm_version
            if not capabilities.check_evm_version(compiler_evm_version):
                msg = "EVM version mismatch! tried to deploy "
                msg += f"{compiler_evm_version} but network only has "
                msg += f"{capabilities.describe_capabilities()}"
                raise Exception(msg)

    @cached_property
    def deployer(self):
        return VyperDeployer(self.compiler_data, filename=self.filename)

    @cached_property
    def abi(self):
        return build_abi_output(self.compiler_data)

    @cached_property
    def _constants(self):
        return ConstantsModel(self.compiler_data)


# create a blueprint for use with `create_from_blueprint`.
# uses a ERC5202 preamble, when calling `create_from_blueprint` will
# need to use `code_offset=3`
class VyperBlueprint(_BaseVyperContract):
    def __init__(
        self,
        compiler_data,
        env=None,
        override_address=None,
        blueprint_preamble=None,
        contract_name=None,
        filename=None,
        gas=None,
    ):
        super().__init__(compiler_data, contract_name, env, filename)

        deploy_bytecode = generate_blueprint_bytecode(
            compiler_data.bytecode, blueprint_preamble
        )

        addr, computation = self.env.deploy(
            bytecode=deploy_bytecode, override_address=override_address, gas=gas
        )
        if computation.is_error:
            raise computation.error

        self.bytecode = computation.output

        self._address = Address(addr)

        self.env.register_blueprint(compiler_data.bytecode, self)


class FrameDetail(dict):
    def __init__(self, fn_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fn_name = fn_name

    def __repr__(self):
        detail = ", ".join(f"{k}={v}" for (k, v) in self.items())
        return f"<{self.fn_name}: {detail}>"


@dataclass
class DevReason:
    reason_type: str
    reason_str: str

    @classmethod
    def at_source_location(
        cls, source_code: str, lineno: int, end_lineno: int
    ) -> Optional["DevReason"]:
        s = reason_at(source_code, lineno, end_lineno)
        if s is None:
            return None
        reason_type, reason_str = s
        return cls(reason_type, reason_str)

    def __str__(self):
        return f"<{self.reason_type}: {self.reason_str}>"


@dataclass
class ErrorDetail:
    vm_error: VMError
    contract_repr: str  # string representation of the contract for the error
    error_detail: str  # compiler provided error detail
    dev_reason: DevReason
    frame_detail: FrameDetail
    ast_source: vy_ast.VyperNode

    @classmethod
    def from_computation(cls, contract, computation):
        error_detail = contract.find_error_meta(computation)
        ast_source = contract.find_source_of(computation)
        reason = None
        if ast_source is not None:
            reason = DevReason.at_source_location(
                ast_source.full_source_code, ast_source.lineno, ast_source.end_lineno
            )
        frame_detail = contract.debug_frame(computation)

        contract_repr = computation._contract_repr_before_revert or repr(contract)
        return cls(
            vm_error=computation.error,
            contract_repr=contract_repr,
            error_detail=error_detail,
            dev_reason=reason,
            frame_detail=frame_detail,
            ast_source=ast_source,
        )

    @property
    def pretty_vm_reason(self):
        err = self.vm_error
        # decode error msg if it's "Error(string)"
        # b"\x08\xc3y\xa0" == method_id("Error(string)")
        if isinstance(err.args[0], bytes) and err.args[0][:4] == b"\x08\xc3y\xa0":
            return abi_decode("(string)", err.args[0][4:])[0]

        return repr(err)

    def __str__(self):
        msg = f"{self.contract_repr}\n"

        if self.error_detail is not None:
            msg += f" <compiler: {self.error_detail}>"

        if self.ast_source is not None:
            # VyperException.__str__ does a lot of formatting for us
            msg = str(VyperException(msg, self.ast_source))

        if self.frame_detail is not None:
            self.frame_detail.fn_name = "locals"  # override the displayed name
            if len(self.frame_detail) > 0:
                msg += f" {self.frame_detail}"

        return msg


# "pattern match" a BoaError. tries to match fields of the error
# to the args/kwargs provided. raises if no match
def check_boa_error_matches(error, *args, **kwargs):
    assert isinstance(error, BoaError)

    def _check(cond, msg=""):
        if not cond:
            raise ValueError(msg)

    frame = error.stack_trace.last_frame
    if len(args) > 0:
        assert len(args) == 1, "multiple args!"
        assert len(kwargs) == 0, "can't mix args and kwargs!"
        err = args[0]
        if isinstance(frame, str):
            # frame for unknown contracts is a string
            _check(err in frame, f"{frame} does not match {args}")
            return

        # try to match anything
        _check(
            err == frame.pretty_vm_reason
            or err == frame.error_detail
            or (frame.dev_reason and err == frame.dev_reason.reason_str),
            f"does not match {args}",
        )
        return

    # try to match a specific kwarg
    assert len(kwargs) == 1 and len(args) == 0

    if isinstance(frame, str):
        # frame for unknown contracts is a string
        raise ValueError(f"expected {kwargs} but got {frame}")

    # don't accept magic
    if frame.dev_reason:
        assert frame.dev_reason.reason_type not in ("vm_error", "compiler")

    k, v = next(iter(kwargs.items()))
    if k == "compiler":
        _check(v == frame.error_detail, f"{frame.error_detail} != {v}")
    elif k == "vm_error":
        _check(
            frame.error_detail == "user revert with reason"
            and v == frame.pretty_vm_reason,
            f"{frame.pretty_vm_reason} != {v}",
        )
    # assume it is a dev reason string
    else:
        assert_ast_types = (vy_ast.Assert, vy_ast.Raise)
        if frame.ast_source.get_ancestor(assert_ast_types) is not None:
            # if it's a dev reason on an assert statement, check that
            # we are actually handling the user assertion and not some other
            # error_detail.
            _check(
                frame.error_detail in DEV_REASON_ALLOWED,
                f"expected <{k}: {v}> but got <compiler: {frame.error_detail}>",
            )
        _check(
            frame.dev_reason is not None
            and k == frame.dev_reason.reason_type
            and v == frame.dev_reason.reason_str,
            f"expected <{k}: {v}> but got {frame.dev_reason}",
        )


# using sha3 preimages, take a storage key and undo
# hashes to get the sequence of hashes ("path") that gave us this image.
def unwrap_storage_key(sha3_db, k):
    path = []

    def unwrap(k):
        k_bytes = to_bytes(k)
        if k_bytes in sha3_db:
            preimage = sha3_db[k_bytes]
            slot, k = preimage[:32], preimage[32:]

            unwrap(slot)

        path.append(k)

    unwrap(k)
    return path


def setpath(lens, path, val):
    for i, k in enumerate(path):
        if i == len(path) - 1:
            lens[k] = val
        else:
            lens = lens.setdefault(k, {})


class StorageVar:
    def __init__(self, contract, slot, typ):
        self.contract = contract
        self.addr = self.contract._address
        self.slot = slot
        self.typ = typ

    def _decode(self, slot, typ, truncate_limit=None):
        n = typ.memory_bytes_required
        if truncate_limit is not None and n > truncate_limit:
            return None  # indicate failure to caller

        fakemem = ByteAddressableStorage(self.contract.env.evm, self.addr, slot)
        return decode_vyper_object(fakemem, typ)

    def _dealias(self, maybe_address):
        try:
            return self.contract.env.lookup_alias(maybe_address)
        except KeyError:  # not found, return the input
            return maybe_address

    def get(self, truncate_limit=None):
        if isinstance(self.typ, HashMapT):
            ret = {}
            for k in self.contract.env.sstore_trace.get(self.addr, {}):
                path = unwrap_storage_key(self.contract.env.sha3_trace, k)
                if to_int(path[0]) != self.slot:
                    continue

                path = path[1:]  # drop the slot
                path_t = []

                ty = self.typ
                for i, p in enumerate(path):
                    path[i] = decode_vyper_object(memoryview(p), ty.key_type)
                    path_t.append(ty.key_type)
                    ty = ty.value_type

                val = self._decode(k, ty, truncate_limit)

                # set val only if value is nonzero
                if val:
                    # decode aliases as needed/possible
                    dealiased_path = []
                    for p, t in zip(path, path_t):
                        if isinstance(t, AddressT):
                            p = self._dealias(p)
                        dealiased_path.append(p)
                    setpath(ret, dealiased_path, val)

            return ret

        else:
            return self._decode(self.slot, self.typ, truncate_limit)


# data structure to represent the storage variables in a contract
class StorageModel:
    def __init__(self, contract):
        compiler_data = contract.compiler_data
        # TODO: recurse into imported modules
        for k, v in contract.module_t.variables.items():
            is_storage = not (v.is_immutable or v.is_constant or v.is_transient)
            if is_storage:
                slot = compiler_data.storage_layout["storage_layout"][k]["slot"]
                setattr(self, k, StorageVar(contract, slot, v.typ))

    def dump(self):
        ret = FrameDetail("storage")

        for k, v in vars(self).items():
            t = v.get(truncate_limit=1024)
            if t is None:
                t = "<truncated>"  # too large, truncated
            ret[k] = t

        return ret


# data structure to represent the storage variables in a contract
class ImmutablesModel:
    def __init__(self, contract):
        compiler_data = contract.compiler_data
        data_section = memoryview(contract.data_section)
        # TODO: recurse into imported modules
        for k, v in contract.module_t.variables.items():
            if v.is_immutable:  # check that v
                ofst = compiler_data.storage_layout["code_layout"][k]["offset"]
                immutable_raw_bytes = data_section[ofst:]
                value = decode_vyper_object(immutable_raw_bytes, v.typ)
                setattr(self, k, value)

    def dump(self):
        return FrameDetail("immutables", vars(self))

    def __repr__(self):
        return repr(self.dump())


# data structure to represent the constants in a contract
class ConstantsModel:
    def __init__(self, compiler_data: CompilerData):
        for v in compiler_data.annotated_vyper_module.get_children(VariableDecl):
            if v.is_constant:
                setattr(self, v.target.id, v.value.get_folded_value().value)

    def dump(self):
        return FrameDetail("constants", vars(self))

    def __repr__(self):
        return repr(self.dump())


class VyperContract(_BaseVyperContract):
    _can_line_profile = True

    def __init__(
        self,
        compiler_data: CompilerData,
        *args,
        value=0,
        env: Env = None,
        override_address: Address = None,
        # whether to skip constructor
        skip_initcode=False,
        created_from: Address = None,
        contract_name=None,
        filename: str = None,
        gas=None,
    ):
        super().__init__(compiler_data, contract_name, env, filename)

        self.created_from = created_from
        self._computation = None
        self._source_map = None

        # add all exposed functions from the interface to the contract
        exposed_fns = {
            fn_t.name: fn_t.decl_node
            for fn_t in compiler_data.global_ctx.exposed_functions
        }

        # set external methods as class attributes:
        self._ctor = None
        if compiler_data.global_ctx.init_function is not None:
            self._ctor = VyperFunction(
                compiler_data.global_ctx.init_function.decl_node, self
            )

        if skip_initcode:
            if value:
                raise Exception("nonzero value but initcode is being skipped")
            addr = Address(override_address)
        else:
            addr = self._run_init(
                *args, value=value, override_address=override_address, gas=gas
            )
        self._address = addr

        for fn_name, fn in exposed_fns.items():
            setattr(self, fn_name, VyperFunction(fn, self))

        # set internal methods as class.internal attributes:
        self.internal = lambda: None
        for fn in self.module_t.function_defs:
            if not fn._metadata["func_type"].is_internal:
                continue
            setattr(self.internal, fn.name, VyperInternalFunction(fn, self))

        # TODO: set library methods as class.internal attributes?

        # not sure if this is accurate in the presence of modules
        self._function_id = len(self.module_t.function_defs)

        self._storage = StorageModel(self)

        self._eval_cache = lrudict(0x1000)

        self.env.register_contract(self._address, self)

    def _run_init(self, *args, value=0, override_address=None, gas=None):
        encoded_args = b""
        if self._ctor:
            encoded_args = self._ctor.prepare_calldata(*args)

        initcode = self.compiler_data.bytecode + encoded_args
        with self._anchor_source_map(self._deployment_source_map):
            address, computation = self.env.deploy(
                bytecode=initcode,
                value=value,
                override_address=override_address,
                gas=gas,
                contract=self,
            )

            self._computation = computation
            self.bytecode = computation.output

            if computation.is_error:
                self.handle_error(computation)

            return address

    @cached_property
    def _deployment_source_map(self):
        with anchor_settings(self.compiler_data.settings):
            _, source_map = compile_ir.assembly_to_evm(self.compiler_data.assembly)
            return source_map

    # manually set the runtime bytecode, instead of using deploy
    def _set_bytecode(self, bytecode: bytes) -> None:
        to_check = bytecode
        if self.data_section_size != 0:
            to_check = bytecode[: -self.data_section_size]
        assert isinstance(self.compiler_data, CompilerData)
        if to_check != self.compiler_data.bytecode_runtime:
            warnings.warn(
                f"casted bytecode does not match compiled bytecode at {self}",
                stacklevel=2,
            )
        self.bytecode = bytecode

    def __repr__(self):
        ret = (
            f"<{self.compiler_data.contract_path} at {self.address}, "
            f"compiled with vyper-{vyper.__version__}+{vyper.__commit__}>"
        )

        if self.created_from is not None:
            ret += f" (created by {self.created_from})"

        dump_storage = True  # maybe make this configurable in the future
        storage_detail = self._storage.dump()
        if dump_storage and len(storage_detail) > 0:
            ret += f"\n{storage_detail}"

        return ret

    @cached_property
    def _immutables(self):
        return ImmutablesModel(self)

    # is this actually useful?
    def at(self, address):
        return self.deployer.at(address)

    def _get_fn_from_computation(self, computation):
        node = self.find_source_of(computation)
        return get_fn_ancestor_from_node(node)

    def debug_frame(self, computation=None):
        if computation is None:
            computation = self._computation

        fn = self._get_fn_from_computation(computation)
        if fn is None:
            # TODO: figure out why fn is None.
            return None

        fn_t = fn._metadata["func_type"]

        frame_info = fn_t._ir_info.frame_info

        mem = computation._memory
        frame_detail = FrameDetail(fn.name)

        # ensure memory is initialized for `decode_vyper_object()`
        mem.extend(frame_info.frame_start, frame_info.frame_size)
        for k, v in frame_info.frame_vars.items():
            if v.location.name != "memory":
                continue
            ofst = v.pos
            size = v.typ.memory_bytes_required
            frame_detail[k] = decode_vyper_object(mem.read(ofst, size), v.typ)

        return frame_detail

    @property
    def module_t(self):
        return self.compiler_data.global_ctx

    # TODO: maybe rename to `ast_map`
    @property
    def source_map(self):
        if self._source_map is None:
            with anchor_settings(self.compiler_data.settings):
                assembly = self.compiler_data.assembly_runtime
                _, self._source_map = compile_ir.assembly_to_evm(assembly)
        return self._source_map

    def find_error_meta(self, computation):
        if hasattr(computation, "vyper_error_msg"):
            # this is set by ir executor currently.
            return computation.vyper_error_msg

        code_stream = computation.code
        error_map = self.source_map.get("error_map", {})
        for pc in reversed(code_stream._trace):
            if pc in error_map:
                return error_map[pc]
        return None

    def find_source_of(self, computation):
        if hasattr(computation, "vyper_source_pos"):
            # this is set by ir executor currently.
            return self.source_map.get(computation.vyper_source_pos)

        code_stream = computation.code
        ast_map = self.source_map["pc_raw_ast_map"]
        for pc in reversed(code_stream._trace):
            if pc in ast_map:
                return ast_map[pc]
        return None

    def trace_source(self, computation) -> Optional["VyperTraceSource"]:
        if (node := self.find_source_of(computation)) is None:
            return None
        return VyperTraceSource(self, node, method_id=computation.msg.data[:4])

    # ## handling events
    def _get_logs(self, computation, include_child_logs):
        if computation is None:
            return []

        if include_child_logs:
            return list(computation.get_raw_log_entries())

        return computation._log_entries

    def get_logs(self, computation=None, include_child_logs=True):
        if computation is None:
            computation = self._computation

        entries = self._get_logs(computation, include_child_logs)

        # py-evm log format is (log_id, topics, data)
        # sort on log_id
        entries = sorted(entries)

        ret = []
        for e in entries:
            logger_address = e[1]
            c = self.env.lookup_contract(logger_address)
            if c is not None:
                ret.append(c.decode_log(e))
            else:
                ret.append(RawEvent(e))

        return ret

    @cached_property
    def event_for(self):
        module_t = self.compiler_data.global_ctx
        return {e.event_id: e for e in module_t.used_events}

    def decode_log(self, e):
        log_id, address, topics, data = e
        assert self._address.canonical_address == address
        event_hash = topics[0]
        event_t = self.event_for[event_hash]

        topic_typs = []
        arg_typs = []
        for is_topic, typ in zip(event_t.indexed, event_t.arguments.values()):
            if not is_topic:
                arg_typs.append(typ)
            else:
                topic_typs.append(typ)

        decoded_topics = []
        for typ, t in zip(topic_typs, topics[1:]):
            # convert to bytes for abi decoder
            encoded_topic = t.to_bytes(32, "big")
            decoded_topics.append(
                abi_decode(typ.abi_type.selector_name(), encoded_topic)
            )

        tuple_typ = TupleT(arg_typs)

        args = abi_decode(tuple_typ.abi_type.selector_name(), data)

        return Event(log_id, self._address, event_t, decoded_topics, args)

    def marshal_to_python(self, computation, vyper_typ):
        self._computation = computation  # for further inspection

        if computation.is_error:
            self.handle_error(computation)

        # cache gas used for call if profiling is enabled
        gas_meter = self.env.get_gas_meter_class()
        if gas_meter == ProfilingGasMeter:
            cache_gas_used_for_computation(self, computation)

        if vyper_typ is None:
            return None

        # selfdestruct
        if len(computation.beneficiaries) > 0:
            return None

        return_typ = calculate_type_for_external_return(vyper_typ)
        ret = abi_decode(return_typ.abi_type.selector_name(), computation.output)

        # unwrap the tuple if needed
        if not isinstance(vyper_typ, TupleT):
            (ret,) = ret

        return vyper_object(ret, vyper_typ)

    def stack_trace(self, computation=None):
        computation = computation or self._computation
        ret = StackTrace([ErrorDetail.from_computation(self, computation)])
        error_detail = self.find_error_meta(computation)
        if error_detail not in EXTERNAL_CALL_ERRORS + CREATE_ERRORS:
            return ret
        return _handle_child_trace(computation, self.env, ret)

    def ensure_id(self, fn_t):  # mimic vyper.codegen.module.IDGenerator api
        if fn_t._function_id is None:
            fn_t._function_id = self._function_id
            self._function_id += 1

    @cached_property
    def _vyper_namespace(self):
        module = self.compiler_data.annotated_vyper_module
        # make a copy of the namespace, since we might modify it
        ret = copy.copy(module._metadata["namespace"])
        ret._scopes = copy.deepcopy(ret._scopes)
        if len(ret._scopes) == 0:
            # funky behavior in Namespace.enter_scope()
            ret._scopes.append(set())
        return ret

    @contextlib.contextmanager
    def override_vyper_namespace(self):
        # ensure self._vyper_namespace is computed
        contract_members = self._vyper_namespace["self"].typ.members
        try:
            to_keep = set(contract_members.keys())
            with vy_ns.override_global_namespace(self._vyper_namespace):
                yield
        finally:
            # drop all keys which were added while yielding
            keys = list(contract_members.keys())
            for k in keys:
                if k not in to_keep:
                    contract_members.pop(k)

    # for eval(), we need unoptimized assembly, since the dead code
    # eliminator might prune a dead function (which we want to eval)
    @cached_property
    def unoptimized_assembly(self):
        with anchor_settings(self.compiler_data.settings):
            runtime = self.unoptimized_ir[1]
            return compile_ir.compile_to_assembly(
                runtime, optimize=OptimizationLevel.NONE
            )

    @cached_property
    def data_section_size(self):
        return self.module_t.immutable_section_bytes

    @cached_property
    def data_section(self):
        # extract the data section from the bytecode
        if self.data_section_size:
            return self.bytecode[-self.data_section_size :]
        else:
            return b""

    @cached_property
    def unoptimized_bytecode(self):
        with anchor_settings(self.compiler_data.settings):
            s, _ = compile_ir.assembly_to_evm(
                self.unoptimized_assembly, insert_vyper_signature=True
            )
            return s + self.data_section

    @cached_property
    def unoptimized_ir(self):
        settings = copy.copy(self.compiler_data.settings)
        settings.optimize = OptimizationLevel.NONE
        with anchor_settings(settings):
            return generate_ir_for_module(self.module_t)

    @cached_property
    def ir_executor(self):
        _, ir_runtime = self.unoptimized_ir
        with anchor_settings(self.compiler_data.settings):
            return executor_from_ir(ir_runtime, self.compiler_data)

    @contextlib.contextmanager
    def _anchor_source_map(self, source_map):
        tmp = self._source_map
        try:
            self._source_map = source_map
            yield
        finally:
            self._source_map = tmp

    def eval(
        self,
        stmt: str,
        value: int = 0,
        gas: Optional[int] = None,
        sender: Optional[Address] = None,
    ) -> Any:
        """eval vyper code in the context of this contract"""

        # this method is super slow so we cache compilation results
        if stmt not in self._eval_cache:
            self._eval_cache[stmt] = generate_bytecode_for_arbitrary_stmt(stmt, self)
        _, ir_executor, bytecode, source_map, typ = self._eval_cache[stmt]

        with self._anchor_source_map(source_map):
            method_id = b"dbug"  # note dummy method id, doesn't get validated
            c = self.env.execute_code(
                to_address=self._address,
                sender=sender,
                data=method_id,
                value=value,
                gas=gas,
                contract=self,
                override_bytecode=bytecode,
                ir_executor=ir_executor,
            )

            return self.marshal_to_python(c, typ)

    # inject a function into this VyperContract without affecting the
    # contract's source code. useful for testing private functionality
    def inject_function(self, fn_source_code, force=False):
        if not hasattr(self, "inject"):
            self.inject = lambda: None

        # get an AST so we know the fn name; work is doubled in
        # _compile_vyper_function but no way around it.
        fn_ast = parse_to_ast(fn_source_code).body[0]
        if hasattr(self.inject, fn_ast.name) and not force:
            raise ValueError(f"already injected: {fn_ast.name}")

        # ensure self._vyper_namespace is computed
        self._vyper_namespace["self"].typ.members.pop(fn_ast.name, None)
        f = _InjectVyperFunction(self, fn_source_code)
        setattr(self.inject, fn_ast.name, f)


class VyperFunction:
    def __init__(self, fn_ast, contract):
        super().__init__()
        self.fn_ast = fn_ast
        self.contract = contract
        self.env = contract.env

        self.__doc__ = (
            fn_ast.doc_string.value if hasattr(fn_ast, "doc_string") else None
        )
        self.__module__ = self.contract.compiler_data.contract_path

    def __repr__(self):
        return f"{self.contract.compiler_data.contract_path}.{self.fn_ast.name}"

    def __str__(self):
        return repr(self.func_t)

    @cached_property
    def _source_map(self):
        return self.contract.source_map

    @property
    def func_t(self):
        return self.fn_ast._metadata["func_type"]

    @cached_property
    def ir(self):
        module_t = self.contract.module_t

        if self.func_t.is_internal:
            res = generate_ir_for_internal_function(self.fn_ast, module_t, False)
            ir = res.func_ir
        else:
            res = generate_ir_for_external_function(self.fn_ast, module_t)
            ir = res.common_ir

        return optimize(ir)

    @cached_property
    def assembly(self):
        ir = IRnode.from_list(
            ["with", _METHOD_ID_VAR, ["shr", 224, ["calldataload", 0]], self.ir]
        )
        return compile_ir.compile_to_assembly(ir)

    @cached_property
    def opcodes(self):
        return compiler_output._build_opcodes(self.bytecode)

    @cached_property
    def bytecode(self):
        bytecode, _ = compile_ir.assembly_to_evm(self.assembly)
        return bytecode

    # hotspot, cache the signature computation
    def args_abi_type(self, num_kwargs):
        if not hasattr(self, "_signature_cache"):
            self._signature_cache = {}

        if num_kwargs in self._signature_cache:
            return self._signature_cache[num_kwargs]

        # align the kwargs with the signature
        sig_kwargs = self.func_t.keyword_args[:num_kwargs]
        sig_args = self.func_t.positional_args + sig_kwargs
        args_abi_type = (
            "(" + ",".join(arg.typ.abi_type.selector_name() for arg in sig_args) + ")"
        )
        abi_sig = self.func_t.name + args_abi_type

        _method_id = method_id(abi_sig)
        self._signature_cache[num_kwargs] = (_method_id, args_abi_type)

        return _method_id, args_abi_type

    def prepare_calldata(self, *args, **kwargs):
        n_total_args = self.func_t.n_total_args
        n_pos_args = self.func_t.n_positional_args

        if not n_pos_args <= len(args) <= n_total_args:
            expectation_str = f"expected between {n_pos_args} and {n_total_args}"
            if n_pos_args == n_total_args:
                expectation_str = f"expected {n_total_args}"
            raise Exception(
                f"bad args to `{repr(self.func_t)}` "
                f"({expectation_str}, got {len(args)})"
            )

        # align the kwargs with the signature
        # sig_kwargs = self.func_t.default_args[: len(kwargs)]

        total_non_base_args = len(kwargs) + len(args) - n_pos_args

        args = [getattr(arg, "address", arg) for arg in args]

        method_id, args_abi_type = self.args_abi_type(total_non_base_args)
        encoded_args = abi_encode(args_abi_type, args)

        if self.func_t.is_constructor or self.func_t.is_fallback:
            return encoded_args

        return method_id + encoded_args

    def __call__(self, *args, value=0, gas=None, sender=None, **kwargs):
        calldata_bytes = self.prepare_calldata(*args, **kwargs)

        # getattr(x, attr, None) swallows exceptions. use explicit hasattr+getattr
        ir_executor = None
        if hasattr(self, "_ir_executor"):
            ir_executor = self._ir_executor

        override_bytecode = None
        if hasattr(self, "_override_bytecode"):
            override_bytecode = self._override_bytecode

        # note: this anchor doesn't do anything on the default implementation.
        # the source map is overridden in subclasses
        with self.contract._anchor_source_map(self._source_map):
            computation = self.env.execute_code(
                to_address=self.contract._address,
                sender=sender,
                data=calldata_bytes,
                value=value,
                gas=gas,
                is_modifying=self.func_t.is_mutable,
                override_bytecode=override_bytecode,
                ir_executor=ir_executor,
                contract=self.contract,
            )

            typ = self.func_t.return_type
            return self.contract.marshal_to_python(computation, typ)


class VyperInternalFunction(VyperFunction):
    """Internal contract functions are exposed by wrapping it with a dummy
    external contract function, appending the wrapper's ast at the top of
    the contract and then generating bytecode to run internal methods
    (as external methods).
    """

    @cached_property
    def _compiled(self):
        return generate_bytecode_for_internal_fn(self)

    # OVERRIDE so that __call__ uses the specially crafted bytecode
    @cached_property
    def _override_bytecode(self):
        _, _, bytecode, _, _ = self._compiled
        return bytecode

    @cached_property
    def _ir_executor(self):
        _, ir_executor, _, _, _ = self._compiled
        return ir_executor

    # OVERRIDE so that __call__ uses corresponding source map
    @cached_property
    def _source_map(self):
        _, _, _, source_map, _ = self._compiled
        return source_map


class VyperTraceSource(TraceSource):
    def __init__(
        self, contract: VyperContract, node: vy_ast.VyperNode, method_id: bytes
    ):
        self.contract = contract
        self.node = node
        self.method_id = method_id

    def __str__(self):
        return f"{self.contract.contract_name}.{self.func_t.name}:{self.node.lineno}"

    def __repr__(self):
        return repr(self.node)

    @cached_property
    def _func_t_helper(self):
        method_id_int = int(self.method_id.hex(), 16)
        for fn_t in self.contract.compiler_data.global_ctx.exposed_functions:
            for schema, id_ in fn_t.method_ids.items():
                if id_ == method_id_int:
                    return schema, fn_t

    @property
    def func_t(self):
        return self._func_t_helper[1]

    @cached_property
    def args_abi_type(self) -> str:
        schema, fn_t = self._func_t_helper
        return schema.replace(f"{fn_t.name}(", "(")

    @cached_property
    def _argument_names(self) -> list[str]:
        return [arg.name for arg in self.func_t.arguments]

    @cached_property
    def return_abi_type(self) -> str:  # must be implemented by subclasses
        typ = self.func_t.return_type
        if typ is None:
            return "()"
        return typ.abi_type.selector_name()


class _InjectVyperFunction(VyperFunction):
    def __init__(self, contract, fn_source):
        ast, ir_executor, bytecode, source_map, _ = compile_vyper_function(
            fn_source, contract
        )
        super().__init__(ast, contract)

        # OVERRIDES so that __call__ does the right thing
        self._override_bytecode = bytecode
        self._ir_executor = ir_executor
        self._source_map = source_map


_typ_cache = {}


def vyper_object(val, vyper_type):
    # make a thin wrapper around whatever type val is,
    # and tag it with _vyper_type metadata

    vt = type(val)
    if vt is bool or vt is Address:
        # https://stackoverflow.com/q/2172189
        # bool is not ambiguous wrt vyper type anyways.
        return val

    if vt not in _typ_cache:
        # ex. class int_wrapper(int): pass
        _typ_cache[vt] = type(f"{vt.__name__}_wrapper", (vt,), {})

    t = _typ_cache[type(val)]

    ret = t(val)
    ret._vyper_type = vyper_type
    return ret



import time
from typing import Any
from urllib.parse import urlparse

import requests

try:
    import ujson as json
except ImportError:  # pragma: no cover
    import json  # type: ignore

TIMEOUT = 60  # default timeout for http requests in seconds


# some utility functions


def trim_dict(kv):
    return {k: v for (k, v) in kv.items() if bool(v)}


def fixup_dict(kv):
    return {k: to_hex(v) for (k, v) in trim_dict(kv).items()}


def to_hex(s: int | bytes | str) -> str:
    if isinstance(s, int):
        return hex(s)
    if isinstance(s, bytes):
        return "0x" + s.hex()
    if isinstance(s, str):
        assert s.startswith("0x")
        return s
    raise TypeError(
        f"to_hex expects bytes, int or (hex) string, but got {type(s)}: {s}"
    )


def to_int(hex_str: str) -> int:
    if hex_str == "0x":
        return 0
    return int(hex_str, 16)


def to_bytes(hex_str: str) -> bytes:
    return bytes.fromhex(hex_str.removeprefix("0x"))


class RPCError(Exception):
    def __init__(self, message: str, code: int):
        super().__init__(f"{code}: {message}")
        self.code = code

    @classmethod
    def from_json(cls, data):
        return cls(message=data["message"], code=data["code"])


class RPC:
    """
    Base class for RPC implementations.
    This abstract class does not use ABC for performance reasons.
    """

    @property
    def identifier(self) -> str:  # pragma: no cover
        raise NotImplementedError

    @property
    def name(self) -> str:  # pragma: no cover
        raise NotImplementedError

    def fetch_uncached(self, method, params):
        return self.fetch(method, params)

    def fetch(self, method: str, params: Any) -> Any:  # pragma: no cover
        raise NotImplementedError

    def fetch_multi(
        self, payloads: list[tuple[str, Any]]
    ) -> list[Any]:  # pragma: no cover
        raise NotImplementedError

    def wait_for_tx_receipt(self, tx_hash, timeout: float, poll_latency=0.25):
        start = time.time()

        while True:
            receipt = self.fetch_uncached("eth_getTransactionReceipt", [tx_hash])
            if receipt is not None:
                return receipt
            if time.time() + poll_latency > start + timeout:
                raise ValueError(f"Timed out waiting for ({tx_hash})")
            time.sleep(poll_latency)


class EthereumRPC(RPC):
    def __init__(self, url: str):
        self._rpc_url = url
        self._session = requests.Session()

        # declare app name to frame.sh
        self._session.headers["Origin"] = "Titanoboa"

    @property
    def identifier(self):
        return self._rpc_url

    @property
    def name(self):
        # return a version of the URL which has everything past the "base"
        # url stripped out (content which you might not want to end up
        # in logs)
        parse_result = urlparse(self._rpc_url)
        partial_ret = f"{parse_result.scheme}://{parse_result.netloc}"
        if partial_ret != self._rpc_url:
            return f"{partial_ret} (URL partially masked for privacy)"
        return self._rpc_url

    def fetch(self, method, params):
        # the obvious thing to do here is dispatch into fetch_multi.
        # but some providers (alchemy) can't handle batched requests
        # for certain endpoints (debug_traceTransaction).
        req = {"jsonrpc": "2.0", "method": method, "params": params, "id": 0}
        # print(req)
        res = self._session.post(self._rpc_url, json=req, timeout=TIMEOUT)
        res.raise_for_status()
        res = json.loads(res.text)
        # print(res)
        if "error" in res:
            raise RPCError.from_json(res["error"])
        return res["result"]

    def fetch_multi(self, payloads):
        request = [
            {"jsonrpc": "2.0", "method": method, "params": params, "id": i}
            for i, (method, params) in enumerate(payloads)
        ]
        response = self._session.post(self._rpc_url, json=request, timeout=TIMEOUT)
        response.raise_for_status()

        results = {}  # keep results in a dict to preserve order
        for item in json.loads(response.text):
            if "error" in item:
                raise RPCError.from_json(item["error"])
            results[item["id"]] = item["result"]

        return [results[i] for i in range(len(payloads))]



from collections import defaultdict
from functools import cached_property
from typing import Any, Optional, Union
from warnings import warn

from eth.abc import ComputationAPI
from vyper.semantics.analysis.base import FunctionVisibility, StateMutability
from vyper.utils import method_id

from boa.contracts.base_evm_contract import (
    BoaError,
    StackTrace,
    _BaseEVMContract,
    _handle_child_trace,
)
from boa.contracts.call_trace import TraceSource
from boa.util.abi import ABIError, Address, abi_decode, abi_encode, is_abi_encodable


class ABIFunction:
    """A single function in an ABI. It does not include overloads."""

    def __init__(self, abi: dict, contract_name: str):
        """
        :param abi: the ABI entry for this function
        :param contract_name: the name of the contract this function belongs to
        """
        self._abi = abi
        self._contract_name = contract_name
        self._function_visibility = FunctionVisibility.EXTERNAL
        self._mutability = StateMutability.from_abi(abi)
        self.contract: Optional["ABIContract"] = None

    @property
    def name(self) -> str | None:
        if self.is_constructor:
            return None
        return self._abi["name"]

    @cached_property
    def argument_types(self) -> list:
        return [_abi_from_json(i) for i in self._abi["inputs"]]

    @property
    def argument_count(self) -> int:
        return len(self.argument_types)

    @property
    def signature(self) -> str:
        return f"({_format_abi_type(self.argument_types)})"

    @cached_property
    def return_type(self) -> list:
        return [_abi_from_json(o) for o in self._abi["outputs"]]

    @property
    def full_signature(self) -> str:
        assert self.name is not None, "Constructor does not have a name."
        return f"{self.name}{self.signature}"

    @property
    def pretty_signature(self) -> str:
        return f"{self.pretty_name}{self.signature} -> {self.return_type}"

    @cached_property
    def pretty_name(self):
        if self.is_constructor:
            return "constructor"
        return self.name

    @cached_property
    def method_id(self) -> bytes:
        assert self.name, "Constructor does not have a method id."
        return method_id(self.name + self.signature)

    @cached_property
    def is_constructor(self):
        return self._abi["type"] == "constructor"

    def __repr__(self) -> str:
        return f"ABI {self._contract_name}.{self.pretty_signature}"

    def __str__(self) -> str:
        return repr(self)

    @property
    def is_mutable(self) -> bool:
        return self._mutability > StateMutability.VIEW

    def is_encodable(self, *args, **kwargs) -> bool:
        """Check whether this function accepts the given arguments after eventual encoding."""
        if len(kwargs) + len(args) != self.argument_count:
            return False
        parsed_args = self._merge_kwargs(*args, **kwargs)
        return all(
            is_abi_encodable(abi_type, arg)
            for abi_type, arg in zip(self.argument_types, parsed_args)
        )

    def prepare_calldata(self, *args, **kwargs) -> bytes:
        """Prepare the call data for the function call."""
        abi_args = self._merge_kwargs(*args, **kwargs)
        encoded_args = abi_encode(self.signature, abi_args)
        if self.is_constructor:
            return encoded_args
        return self.method_id + encoded_args

    def _merge_kwargs(self, *args, **kwargs) -> list:
        """Merge positional and keyword arguments into a single list."""
        if len(kwargs) + len(args) != self.argument_count:
            raise TypeError(
                f"Bad args to `{repr(self)}` (expected {self.argument_count} "
                f"arguments, got {len(args)} args and {len(kwargs)} kwargs)"
            )
        try:
            kwarg_inputs = self._abi["inputs"][len(args) :]
            return list(args) + [kwargs.pop(i["name"]) for i in kwarg_inputs]
        except KeyError as e:
            error = f"Missing keyword argument {e} for `{self.signature}`. Passed {args} {kwargs}"
            raise TypeError(error)

    def __call__(self, *args, value=0, gas=None, sender=None, **kwargs):
        """Calls the function with the given arguments based on the ABI contract."""
        if not self.contract or not self.contract.env:
            raise Exception(f"Cannot call {self} without deploying contract.")

        computation = self.contract.env.execute_code(
            to_address=self.contract.address,
            sender=sender,
            data=self.prepare_calldata(*args, **kwargs),
            value=value,
            gas=gas,
            is_modifying=self.is_mutable,
            contract=self.contract,
        )

        match self.contract.marshal_to_python(computation, self.return_type):
            case ():
                return None
            case (single,):
                return single
            case multiple:
                return tuple(multiple)


class ABIOverload:
    """
    Represents a set of functions that have the same name but different
    argument types. This is used to implement function overloading.
    """

    @staticmethod
    def create(
        functions: list[ABIFunction], contract: "ABIContract"
    ) -> Union["ABIOverload", ABIFunction]:
        """
        Create an ABIOverload if there are multiple functions, otherwise
        return the single function.
        :param functions: a list of functions with the same name
        :param contract: the ABIContract that these functions belong to
        """
        for f in functions:
            f.contract = contract
        if len(functions) == 1:
            return functions[0]
        return ABIOverload(functions)

    def __init__(self, functions: list[ABIFunction]):
        self.functions = functions

    @cached_property
    def name(self) -> str | None:
        return self.functions[0].name

    def prepare_calldata(self, *args, disambiguate_signature=None, **kwargs) -> bytes:
        """Prepare the calldata for the function that matches the given arguments."""
        function = self._pick_overload(
            *args, disambiguate_signature=disambiguate_signature, **kwargs
        )
        return function.prepare_calldata(*args, **kwargs)

    def __call__(
        self,
        *args,
        value=0,
        gas=None,
        sender=None,
        disambiguate_signature=None,
        **kwargs,
    ):
        """
        Call the function that matches the given arguments.
        :raises Exception: if a single function is not found
        """
        function = self._pick_overload(
            *args, disambiguate_signature=disambiguate_signature, **kwargs
        )
        return function(*args, value=value, gas=gas, sender=sender, **kwargs)

    def _pick_overload(
        self, *args, disambiguate_signature=None, **kwargs
    ) -> ABIFunction:
        """Pick the function that matches the given arguments."""
        if disambiguate_signature is None:
            matches = [f for f in self.functions if f.is_encodable(*args, **kwargs)]
        else:
            matches = [
                f for f in self.functions if disambiguate_signature == f.full_signature
            ]
            assert len(matches) <= 1, "ABI signature must be unique"

        assert self.name, "Constructor does not have a name."
        match matches:
            case [function]:
                return function
            case []:
                raise Exception(
                    f"Could not find matching {self.name} function for given arguments."
                )
            case multiple:
                raise Exception(
                    f"Ambiguous call to {self.name}. "
                    f"Arguments can be encoded to multiple overloads: "
                    f"{', '.join(self.name + f.signature for f in multiple)}. "
                    f"(Hint: try using `disambiguate_signature=` to disambiguate)."
                )


class ABIContract(_BaseEVMContract):
    """A deployed contract loaded via an ABI."""

    def __init__(
        self,
        name: str,
        abi: list[dict],
        functions: list[ABIFunction],
        address: Address,
        filename: Optional[str] = None,
        env=None,
    ):
        super().__init__(name, env, filename=filename, address=address)
        self._abi = abi
        self._functions = functions

        self._bytecode = self.env.get_code(address)
        if not self._bytecode:
            warn(
                f"Requested {self} but there is no bytecode at that address!",
                stacklevel=2,
            )

        overloads = defaultdict(list)
        for f in self._functions:
            overloads[f.name].append(f)

        for fn_name, group in overloads.items():
            if fn_name is not None:  # constructors have no name
                setattr(self, fn_name, ABIOverload.create(group, self))

        self._address = Address(address)
        self._computation: Optional[ComputationAPI] = None

    @property
    def abi(self):
        return self._abi

    @cached_property
    def method_id_map(self):
        """
        Returns a mapping from method id to function object.
        This is used to create the stack trace when an error occurs.
        """
        return {
            function.method_id: function
            for function in self._functions
            if not function.is_constructor
        }

    def marshal_to_python(self, computation, abi_type: list[str]) -> tuple[Any, ...]:
        """
        Convert the output of a contract call to a Python object.
        :param computation: the computation object returned by `execute_code`
        :param abi_type: the ABI type of the return value.
        """
        self._computation = computation
        if computation.is_error:
            return self.handle_error(computation)

        schema = f"({_format_abi_type(abi_type)})"
        try:
            return abi_decode(schema, computation.output)
        except ABIError as e:
            # TODO: the likely error here is that no code exists at the address,
            # it might be better to just let the raw ABIError float up
            raise BoaError.create(computation, self) from e

    def stack_trace(self, computation: ComputationAPI) -> StackTrace:
        """
        Create a stack trace for a failed contract call.
        """
        reason = ""
        if computation.is_error:
            reason = " ".join(str(arg) for arg in computation.error.args if arg != b"")

        calldata_method_id = bytes(computation.msg.data[:4])
        if calldata_method_id in self.method_id_map:
            function = self.method_id_map[calldata_method_id]
            msg = f"  {reason}({self}.{function.pretty_signature})"
        else:
            # Method might not be specified in the ABI
            msg = f"  {reason}(unknown method id {self}.0x{calldata_method_id.hex()})"

        return_trace = StackTrace([msg])
        return _handle_child_trace(computation, self.env, return_trace)

    def trace_source(self, computation) -> Optional["ABITraceSource"]:
        """
        Find the source of the error in the contract.
        :param computation: the computation object returned by `execute_code`
        """
        method_id_ = computation.msg.data[:4]
        if method_id_ not in self.method_id_map:
            return None
        return ABITraceSource(self, self.method_id_map[method_id_])

    @property
    def deployer(self) -> "ABIContractFactory":
        """
        Returns a factory that can be used to retrieve another deployed contract.
        """
        return ABIContractFactory(self.contract_name, self._abi, filename=self.filename)

    def __repr__(self):
        file_str = f" (file {self.filename})" if self.filename else ""
        warn_str = "" if self._bytecode else " (WARNING: no bytecode at this address!)"
        return f"<{self.contract_name} interface at {self.address}{warn_str}>{file_str}"


class ABIContractFactory:
    """
    Represents an ABI contract that has not been coupled with an address yet.
    This is named `Factory` instead of `Deployer` because it doesn't actually
    do any contract deployment.
    """

    def __init__(self, name: str, abi: list[dict], filename: Optional[str] = None):
        self._name = name
        self._abi = abi
        self.filename = filename

    @cached_property
    def abi(self):
        return self._abi

    @property
    def functions(self):
        return [
            ABIFunction(item, self._name)
            for item in self.abi
            if item.get("type") == "function"
        ]

    @classmethod
    def from_abi_dict(cls, abi, name="<anonymous contract>", filename=None):
        return cls(name, abi, filename)

    def at(self, address: Address | str) -> ABIContract:
        """
        Create an ABI contract object for a deployed contract at `address`.
        """
        address = Address(address)
        contract = ABIContract(
            self._name, self._abi, self.functions, address, self.filename
        )

        contract.env.register_contract(address, contract)
        return contract


class ABITraceSource(TraceSource):
    def __init__(self, contract: ABIContract, function: ABIFunction):
        self.contract = contract
        self.function = function

    def __str__(self):
        return f"{self.contract.contract_name}.{self.function.pretty_name}"

    def __repr__(self):
        return repr(self.function)

    @cached_property
    def args_abi_type(self):
        return f"({_format_abi_type(self.function.argument_types)})"

    @cached_property
    def _argument_names(self) -> list[str]:
        return [arg["name"] for arg in self.function._abi["inputs"]]

    @cached_property
    def return_abi_type(self):
        return f"({_format_abi_type(self.function.return_type)})"


def _abi_from_json(abi: dict) -> str:
    """
    Parses an ABI type into its schema string.
    :param abi: The ABI type to parse.
    :return: The schema string for the given abi type.
    """
    if "components" in abi:
        components = ",".join([_abi_from_json(item) for item in abi["components"]])
        if abi["type"].startswith("tuple"):
            return f"({components}){abi['type'][5:]}"
        raise ValueError("Components found in non-tuple type " + abi["type"])

    return abi["type"]


def _format_abi_type(types: list) -> str:
    """
    Converts a list of ABI types into a comma-separated string.
    """
    return ",".join(
        item if isinstance(item, str) else f"({_format_abi_type(item)})"
        for item in types
    )



"""
The entry point for managing the execution environment.
"""
# the main "entry point" for patching py-evm.
# handles low level details around state and py-evm tracing.

import contextlib
import random
import warnings
from typing import Any, Optional, TypeAlias

import eth.constants as constants
from eth_typing import Address as PYEVM_Address  # it's just bytes.

from boa.rpc import RPC, EthereumRPC
from boa.util.abi import Address
from boa.vm.gas_meters import GasMeter, NoGasMeter, ProfilingGasMeter
from boa.vm.py_evm import PyEVM

# make mypy happy
_AddressType: TypeAlias = Address | str | bytes | PYEVM_Address


# wrapper class around py-evm which provides a "contract-centric" API
class Env:
    _singleton = None
    _random = random.Random("titanoboa")  # something reproducible
    _coverage_enabled = False

    def __init__(self, fork_try_prefetch_state=False, fast_mode_enabled=False):
        self._gas_price = None

        self._aliases = {}

        # TODO differentiate between origin and sender
        self.eoa = self.generate_address("eoa")

        self._contracts = {}
        self._code_registry = {}

        self.sha3_trace: dict = {}
        self.sstore_trace: dict = {}

        self._gas_tracker = 0

        self.nickname = "pyevm"

        self.evm = PyEVM(self, fast_mode_enabled, fork_try_prefetch_state)

    def set_random_seed(self, seed=None):
        self._random = random.Random(seed)

    def get_gas_price(self):
        return self._gas_price or 0

    def enable_fast_mode(self, flag: bool = True):
        self.evm.enable_fast_mode(flag)

    def fork(
        self,
        url: str,
        reset_traces=True,
        block_identifier="safe",
        deprecated=True,
        **kwargs,
    ):
        if deprecated:
            warnings.warn("using boa.env.fork directly is deprecated; use `boa.fork`!")
        return self.fork_rpc(EthereumRPC(url), reset_traces, block_identifier, **kwargs)

    def fork_rpc(self, rpc: RPC, reset_traces=True, block_identifier="safe", **kwargs):
        """
        Fork the environment to a local chain.
        :param rpc: RPC to fork from
        :param reset_traces: Reset the traces
        :param block_identifier: Block identifier to fork from
        :param kwargs: Additional arguments for the RPC
        """
        # we usually want to reset the trace data structures
        # but sometimes don't, give caller the option.
        if reset_traces:
            self.sha3_trace = {}
            self.sstore_trace = {}

        self.evm.fork_rpc(rpc, block_identifier, **kwargs)

    def get_gas_meter_class(self):
        return self.evm.get_gas_meter_class()

    def set_gas_meter_class(self, cls: type) -> None:
        self.evm.set_gas_meter_class(cls)

    @contextlib.contextmanager
    def gas_meter_class(self, cls):
        tmp = self.evm.get_gas_meter_class()
        try:
            self.set_gas_meter_class(cls)
            yield
        finally:
            self.set_gas_meter_class(tmp)

    def enable_gas_profiling(self) -> None:
        self.set_gas_meter_class(ProfilingGasMeter)

    def disable_gas_metering(self) -> None:
        self.set_gas_meter_class(NoGasMeter)

    def reset_gas_metering_behavior(self) -> None:
        # Reset gas metering to the default behavior
        self.set_gas_meter_class(GasMeter)

    # set balance of address in py-evm
    def set_balance(self, addr, value):
        self.evm.set_balance(Address(addr), value)

    # get balance of address in py-evm
    def get_balance(self, addr):
        return self.evm.get_balance(Address(addr))

    def register_contract(self, address, obj):
        addr = Address(address)
        self._contracts[addr.canonical_address] = obj

        # also register it in the registry for
        # create_minimal_proxy_to and create_copy_of
        bytecode = self.evm.get_code(addr)
        self._code_registry[bytecode] = obj

    def register_blueprint(self, bytecode, obj):
        self._code_registry[bytecode] = obj

    def _lookup_contract_fast(self, address: PYEVM_Address):
        return self._contracts.get(address)

    def lookup_contract(self, address: _AddressType):
        if address == b"":
            return None
        return self._contracts.get(Address(address).canonical_address)

    def alias(self, address, name):
        self._aliases[Address(address).canonical_address] = name

    def lookup_alias(self, address):
        return self._aliases[Address(address).canonical_address]

    # advanced: reset warm/cold counters for addresses and storage
    def _reset_access_counters(self):
        self.evm.reset_access_counters()

    def get_gas_used(self):
        return self._gas_tracker

    def reset_gas_used(self):
        self._gas_tracker = 0
        self._reset_access_counters()

    # context manager which snapshots the state and reverts
    # to the snapshot on exiting the with statement
    @contextlib.contextmanager
    def anchor(self):
        snapshot_id = self.evm.snapshot()
        try:
            with self.evm.patch.anchor():
                yield
        finally:
            self.evm.revert(snapshot_id)

    @contextlib.contextmanager
    def sender(self, address):
        tmp = self.eoa
        self.eoa = Address(address)
        try:
            yield
        finally:
            self.eoa = tmp

    def prank(self, *args, **kwargs):
        return self.sender(*args, **kwargs)

    @classmethod
    def get_singleton(cls):
        if cls._singleton is None:
            cls._singleton = cls()
        return cls._singleton

    def generate_address(self, alias: Optional[str] = None) -> _AddressType:
        t = Address(self._random.randbytes(20))
        if alias is not None:
            self.alias(t, alias)

        return t

    # helper fn
    def _get_sender(self, sender=None) -> Address:
        if sender is None:
            sender = self.eoa
        if self.eoa is None:
            raise ValueError(f"{self}.eoa not defined!")
        return Address(sender)

    def _update_gas_used(self, gas_used: int):
        self._gas_tracker += gas_used

    def deploy(
        self,
        sender: Optional[_AddressType] = None,
        gas: Optional[int] = None,
        value: int = 0,
        bytecode: bytes = b"",
        start_pc: int = 0,  # TODO: This isn't used
        # override the target address:
        override_address: Optional[_AddressType] = None,
        # the calling vyper contract
        contract: Any = None,
    ):
        sender = self._get_sender(sender)

        if override_address is None:
            target_address = self.evm.generate_create_address(sender)
        else:
            target_address = Address(override_address)

        origin = sender  # XXX: consider making this parameterizable
        computation = self.evm.deploy_code(
            sender=sender,
            origin=origin,
            target_address=target_address,
            gas=gas,
            gas_price=self.get_gas_price(),
            value=value,
            bytecode=bytecode,
        )

        if self._coverage_enabled:
            self._trace_computation(computation, contract)

        if computation._gas_meter_class != NoGasMeter:
            self._update_gas_used(computation.get_gas_used())

        return target_address, computation

    def deploy_code(self, *args, **kwargs) -> tuple[Address, bytes]:
        address, computation = self.deploy(*args, **kwargs)
        if computation.is_error:
            raise computation.error
        return address, computation.output

    def raw_call(
        self,
        to_address,
        sender: Optional[_AddressType] = None,
        gas: Optional[int] = None,
        value: int = 0,
        data: bytes = b"",
    ):
        # simple wrapper around `execute_code` to help simulate calling
        # a contract from an EOA.
        ret = self.execute_code(
            to_address=to_address, sender=sender, gas=gas, value=value, data=data
        )
        if ret.is_error:
            # differ from execute_code, consumers of execute_code want to get
            # error returned "silently" (not thru exception handling mechanism)
            # whereas users of call() expect the exception to be thrown, just
            # like a regular contract call.
            raise ret.error

        return ret

    def execute_code(
        self,
        to_address: _AddressType = constants.ZERO_ADDRESS,
        sender: Optional[_AddressType] = None,
        gas: Optional[int] = None,
        value: int = 0,
        data: bytes = b"",
        override_bytecode: Optional[bytes] = None,
        ir_executor: Any = None,
        is_modifying: bool = True,
        start_pc: int = 0,
        fake_codesize: Optional[int] = None,
        contract: Any = None,  # the calling VyperContract
    ) -> Any:
        if gas is None:
            gas = self.evm.get_gas_limit()

        sender = self._get_sender(sender)

        to = Address(to_address)

        bytecode = override_bytecode
        if override_bytecode is None:
            bytecode = self.evm.get_code(to)

        is_static = not is_modifying
        ret = self.evm.execute_code(
            sender=sender,
            to=to,
            gas=gas,
            gas_price=self.get_gas_price(),
            value=value,
            bytecode=bytecode,
            data=data,
            is_static=is_static,
            fake_codesize=fake_codesize,
            start_pc=start_pc,
            ir_executor=ir_executor,
            contract=contract,
        )
        if self._coverage_enabled:
            self._trace_computation(ret, contract)

        if ret._gas_meter_class != NoGasMeter:
            self._update_gas_used(ret.get_gas_used())

        return ret

    # trace pcs for coverage sake. dummy function which
    # just issues the right calls to _trace_cov() to get picked
    # up by coverage. bit ugly, but tracer only allows
    # dynamic_source_filename to be set once per (python) function call,
    # so we need to use this in case the pc trace covers multiple files
    def _trace_computation(self, computation, contract=None):
        # perf: don't trace if contract is None
        if contract is not None and hasattr(contract, "source_map"):
            ast_map = contract.source_map["pc_raw_ast_map"]
            seen_pcs = set()
            for pc in computation.code._trace:
                if pc in seen_pcs:
                    continue
                if (node := ast_map.get(pc)) is not None:
                    mod = node.module_node
                    self._trace_cov(mod.resolved_path, node)
                seen_pcs.add(pc)

        for child in computation.children:
            if child.msg.code_address == b"":
                continue
            child_contract = self._lookup_contract_fast(child.msg.code_address)
            self._trace_computation(child, child_contract)

    def _trace_cov(self, filename, node):
        pass

    def get_code(self, address: _AddressType) -> bytes:
        return self.evm.get_code(Address(address))

    def set_code(self, address: _AddressType, code: bytes) -> None:
        self.evm.set_code(Address(address), code)

    def get_storage(self, address: _AddressType, slot: int) -> int:
        return self.evm.get_storage(Address(address), slot)

    def set_storage(self, address: _AddressType, slot: int, value: int) -> None:
        self.evm.set_storage(Address(address), slot, value)

    # function to time travel
    def time_travel(
        self,
        seconds: Optional[int] = None,
        blocks: Optional[int] = None,
        block_delta: int = 12,
    ) -> None:
        if (seconds is None) == (blocks is None):
            raise ValueError("One of seconds or blocks should be set")
        if seconds is not None:
            blocks = seconds // block_delta
        else:
            assert blocks is not None  # mypy hint
            seconds = blocks * block_delta

        self.evm.patch.timestamp += seconds
        self.evm.patch.block_number += blocks



# wrapper module around whatever encoder we are using
from typing import Annotated, Any

from eth.codecs.abi import nodes
from eth.codecs.abi.decoder import Decoder
from eth.codecs.abi.encoder import Encoder
from eth.codecs.abi.exceptions import ABIError
from eth.codecs.abi.nodes import ABITypeNode
from eth.codecs.abi.parser import Parser
from eth_typing import Address as PYEVM_Address
from eth_utils import to_canonical_address, to_checksum_address

from boa.util.lrudict import lrudict

_parsers: dict[str, ABITypeNode] = {}


# inherit from `str` so that users can compare with regular hex string
# addresses
class Address(str):
    # converting between checksum and canonical addresses is a hotspot;
    # this class contains both and caches recently seen conversions
    __slots__ = ("canonical_address",)
    _cache = lrudict(1024)

    canonical_address: Annotated[PYEVM_Address, "canonical address"]

    def __new__(cls, address):
        if isinstance(address, Address):
            return address

        try:
            return cls._cache[address]
        except KeyError:
            pass

        checksum_address = to_checksum_address(address)
        self = super().__new__(cls, checksum_address)
        self.canonical_address = to_canonical_address(address)
        cls._cache[address] = self
        return self

    def __repr__(self):
        checksum_addr = super().__repr__()
        return f"Address({checksum_addr})"


class _ABIEncoder(Encoder):
    """
    Custom encoder that extracts the address from an `Address` object
    and passes the result to the base encoder.
    """

    @classmethod
    def visit_AddressNode(cls, node: nodes.AddressNode, value) -> bytes:
        value = getattr(value, "address", value)
        return super().visit_AddressNode(node, value)


class _ABIDecoder(Decoder):
    """
    Custom decoder that wraps address results into an `Address` object.
    """

    @classmethod
    def visit_AddressNode(
        cls, node: nodes.AddressNode, value: bytes, checksum: bool = True, **kwargs: Any
    ) -> "Address":
        ret = super().visit_AddressNode(node, value)
        return Address(ret)


def _get_parser(schema: str):
    try:
        return _parsers[schema]
    except KeyError:
        _parsers[schema] = (ret := Parser.parse(schema))
        return ret


def abi_encode(schema: str, data: Any) -> bytes:
    return _ABIEncoder.encode(_get_parser(schema), data)


def abi_decode(schema: str, data: bytes) -> Any:
    return _ABIDecoder.decode(_get_parser(schema), data)


def is_abi_encodable(abi_type: str, data: Any) -> bool:
    try:
        abi_encode(abi_type, data)
        return True
    except ABIError:
        return False



import json
import time
from dataclasses import dataclass
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import Optional

import requests

from boa.util.abi import Address
from boa.util.open_ctx import Open

DEFAULT_BLOCKSCOUT_URI = "https://eth.blockscout.com"


@dataclass
class Blockscout:
    """
    Allows users to verify contracts on Blockscout.
    This is independent of Vyper contracts, and can be used to verify any smart contract.
    """

    uri: str = DEFAULT_BLOCKSCOUT_URI
    api_key: Optional[str] = None
    timeout: timedelta = timedelta(minutes=2)
    backoff: timedelta = timedelta(milliseconds=500)
    backoff_factor: float = 1.1
    retry_http_codes: tuple[int, ...] = (
        HTTPStatus.NOT_FOUND,
        HTTPStatus.INTERNAL_SERVER_ERROR,
        HTTPStatus.SERVICE_UNAVAILABLE,
        HTTPStatus.GATEWAY_TIMEOUT,
    )

    def verify(
        self,
        address: Address,
        contract_name: str,
        solc_json: dict,
        license_type: str = None,
        wait: bool = False,
    ) -> Optional["VerificationResult"]:
        """
        Verify the Vyper contract on Blockscout.
        :param address: The address of the contract.
        :param contract_name: The name of the contract.
        :param solc_json: The solc_json output of the Vyper compiler.
        :param license_type: The license to use for the contract. Defaults to "none".
        :param wait: Whether to return a VerificationResult immediately
                     or wait for verification to complete. Defaults to False
        """
        if license_type is None:
            license_type = "none"

        api_key = self.api_key or ""

        url = f"{self.uri}/api/v2/smart-contracts/{address}/"
        url += f"verification/via/vyper-standard-input?apikey={api_key}"
        data = {
            "compiler_version": solc_json["compiler_version"],
            "license_type": license_type,
        }
        files = {
            "files[0]": (
                contract_name,
                json.dumps(solc_json).encode("utf-8"),
                "application/json",
            )
        }

        response = requests.post(url, data=data, files=files)
        response.raise_for_status()
        print(response.json().get("message"))  # usually verification started

        if not wait:
            return VerificationResult(address, self)

        self.wait_for_verification(address)
        return None

    def wait_for_verification(self, address: Address) -> None:
        """
        Waits for the contract to be verified on Blockscout.
        :param address: The address of the contract.
        """
        timeout = datetime.now() + self.timeout
        wait_time = self.backoff
        while datetime.now() < timeout:
            if self.is_verified(address):
                msg = "Contract verified!"
                msg += f" {self.uri}/address/{address}?tab=contract_code"
                print(msg)
                return
            time.sleep(wait_time.total_seconds())
            wait_time *= self.backoff_factor

        raise TimeoutError("Timeout waiting for verification to complete")

    def is_verified(self, address: Address) -> bool:
        api_key = self.api_key or ""
        url = f"{self.uri}/api/v2/smart-contracts/{address}?apikey={api_key}"

        response = requests.get(url)
        if response.status_code in self.retry_http_codes:
            return False
        response.raise_for_status()
        return response.json().get("is_verified", False)


_verifier = Blockscout()


@dataclass
class VerificationResult:
    address: Address
    verifier: Blockscout

    def wait_for_verification(self):
        self.verifier.wait_for_verification(self.address)

    def is_verified(self):
        return self.verifier.is_verified(self.address)


def _set_verifier(verifier):
    global _verifier
    _verifier = verifier


def get_verifier():
    global _verifier
    return _verifier


# TODO: maybe allow like `set_verifier("blockscout", *args, **kwargs)`
def set_verifier(verifier):
    return Open(get_verifier, _set_verifier, verifier)


def get_verification_bundle(contract_like):
    if not hasattr(contract_like, "deployer"):
        return None
    if not hasattr(contract_like.deployer, "solc_json"):
        return None
    return contract_like.deployer.solc_json


# should we also add a `verify_deployment` function?
def verify(
    contract, verifier=None, license_type: str = None, wait=False
) -> VerificationResult:
    """
    Verifies the contract on a block explorer.
    :param contract: The contract to verify.
    :param verifier: The block explorer verifier to use.
                     Defaults to get_verifier().
    :param license_type: Optional license to use for the contract.
    """
    if verifier is None:
        verifier = get_verifier()

    if (bundle := get_verification_bundle(contract)) is None:
        raise ValueError(f"Not a contract! {contract}")

    return verifier.verify(
        address=contract.address,
        solc_json=bundle,
        contract_name=contract.contract_name,
        license_type=license_type,
        wait=wait,
    )
